package com.koutra.nodejava;

import java.io.Serializable;
import java.util.Calendar;
import java.util.Date;

import com.koutra.nodejava.utils.DateHelper;
import com.koutra.nodejava.utils.HttpConstants;

/**
 * Creates a cookie, a small amount of information sent by a servlet to a Web
 * browser, saved by the browser, and later sent back to the server. A cookie's
 * value can uniquely identify a client, so cookies are commonly used for
 * session management.
 * 
 * A cookie has a name, a single value, and optional attributes such as a
 * comment, path and domain qualifiers, a maximum age, and a version number.
 * Some Web browsers have bugs in how they handle the optional attributes, so
 * use them sparingly to improve the interoperability of your servlets.
 * 
 * The servlet sends cookies to the browser by using the HttpResponse#addCookie
 * method, which adds fields to HTTP response headers to send cookies to the
 * browser, one at a time. The browser is expected to support 20 cookies for
 * each Web server, 300 cookies total, and may limit cookie size to 4 KB each.
 * 
 * The browser returns cookies to the servlet by adding fields to HTTP request
 * headers. Cookies can be retrieved from a request by using the
 * HttpRequest#getCookies method. Several cookies might have the same name but
 * different path attributes.
 * 
 * Cookies affect the caching of the Web pages that use them. HTTP 1.0 does not
 * cache pages that use cookies created with this class. This class does not
 * support the cache control defined with HTTP 1.1.
 * 
 * This class supports both the Version 0 (by Netscape) and Version 1 (by RFC
 * 2109) cookie specifications. By default, cookies are created using Version 0
 * to ensure the best interoperability.
 * 
 * @author pftakas
 */
public class Cookie implements Cloneable, Serializable {

	private static final long serialVersionUID = -4113415716854814276L;

	protected String name;
	protected String value;
	protected String comment;
	protected String domain;
	protected int maxAge;
	protected String path;
	protected boolean secure;
	protected int version;
	protected boolean httpOnly;
	
	/**
	 * Constructs a cookie with the specified name and value.
	 * 
	 * The name must conform to RFC 2109. However, vendors may provide a
	 * configuration option that allows cookie names conforming to the original
	 * Netscape Cookie Specification to be accepted.
	 * 
	 * The name of a cookie cannot be changed once the cookie has been created.
	 * 
	 * The value can be anything the server chooses to send. Its value is
	 * probably of interest only to the server. The cookie's value can be
	 * changed after creation with the setValue method.
	 * 
	 * By default, cookies are created according to the Netscape cookie
	 * specification. The version can be changed with the setVersion method.
	 * 
	 * @param name the name of the cookie
	 * @param value the value of the cookie
	 * @throws IllegalArgumentException if the cookie name is null or empty
	 * or contains any illegal characters (for example, a comma, space, or
	 * semicolon) or matches a token reserved for use by the cookie protocol
	 */
	public Cookie(String name, String value) {
		this.name = name;
		this.value = value;
		this.comment = null;
		this.domain = null;
		this.maxAge = -1;
		this.path = null;
		this.secure = false;
		this.version = 0;
		this.httpOnly = false;
	}
	
	/**
	 * Specifies a comment that describes a cookie's purpose. The comment is
	 * useful if the browser presents the cookie to the user. Comments are not
	 * supported by Netscape Version 0 cookies.
	 * 
	 * @param purpose a String specifying the comment to display to the user.
	 */
	public void setComment(String purpose) {
		this.comment = purpose;
	}

	/**
	 * Returns the comment describing the purpose of this cookie, or null if
	 * the cookie has no comment.
	 * 
	 * @return the comment of the cookie, or null if unspecified
	 */
	public String getComment() {
		return comment;
	}

	/**
	 * Specifies the domain within which this cookie should be presented.
	 * 
	 * The form of the domain name is specified by RFC 2109. A domain name
	 * begins with a dot (.example.com) and means that the cookie is visible to
	 * servers in a specified Domain Name System (DNS) zone (for example,
	 * www.example.com, but not a.b.example.com). By default, cookies are only
	 * returned to the server that sent them.
	 * 
	 * @param domain the domain name within which this cookie is visible; form
	 * is according to RFC 2109
	 */
	public void setDomain(String domain) {
		this.domain = domain;
	}

	/**
	 * Gets the domain name of this Cookie.
	 * 
	 * Domain names are formatted according to RFC 2109.
	 * 
	 * @return the domain name of this Cookie
	 */
	public String getDomain() {
		return domain;
	}

	/**
	 * Sets the maximum age in seconds for this Cookie.
	 * 
	 * A positive value indicates that the cookie will expire after that many
	 * seconds have passed. Note that the value is the maximum age when the
	 * cookie will expire, not the cookie's current age.
	 * 
	 * A negative value means that the cookie is not stored persistently and
	 * will be deleted when the Web browser exits. A zero value causes the
	 * cookie to be deleted.
	 * 
	 * @param expiry an integer specifying the maximum age of the cookie in
	 * seconds; if negative, means the cookie is not stored; if zero, deletes
	 * the cookie.
	 */
	public void setMaxAge(int expiry) {
		this.maxAge = expiry;
	}

	/**
	 * Gets the maximum age in seconds of this Cookie.
	 * 
	 * By default, -1 is returned, which indicates that the cookie will persist
	 * until browser shutdown.
	 * 
	 * @return an integer specifying the maximum age of the cookie in seconds;
	 * if negative, means the cookie persists until browser shutdown.
	 */
	public int getMaxAge() {
		return maxAge;
	}

	/**
	 * Specifies a path for the cookie to which the client should return the
	 * cookie.
	 * 
	 * The cookie is visible to all the pages in the directory you specify, and
	 * all the pages in that directory's subdirectories. A cookie's path must
	 * include the servlet that set the cookie, for example, /catalog, which
	 * makes the cookie visible to all directories on the server under /catalog.
	 * 
	 * Consult RFC 2109 (available on the Internet) for more information on
	 * setting path names for cookies.
	 * 
	 * @param uri a String specifying a path
	 */
	public void setPath(String uri) {
		this.path = uri;
	}

	/**
	 * Returns the path on the server to which the browser returns this cookie.
	 * The cookie is visible to all subpaths on the server.
	 * 
	 * @return a String specifying a path that contains a servlet name,
	 * for example, /catalog
	 */
	public String getPath() {
		return path;
	}

	/**
	 * Indicates to the browser whether the cookie should only be sent using a
	 * secure protocol, such as HTTPS or SSL.
	 * 
	 * The default value is false.
	 * 
	 * @param flag if true, sends the cookie from the browser to the server
	 * only when using a secure protocol; if false, sent on any protocol.
	 */
	public void setSecure(boolean flag) {
		this.secure = flag;
	}

	/**
	 * Returns true if the browser is sending cookies only over a secure
	 * protocol, or false if the browser can send cookies using any protocol.
	 * 
	 * @return true if the browser uses a secure protocol, false otherwise.
	 */
	public boolean getSecure() {
		return secure;
	}

	/**
	 * Returns the name of the cookie. The name cannot be changed after
	 * creation.
	 * 
	 * @return the name of the cookie
	 */
	public String getName() {
		return name;
	}

	/**
	 * Assigns a new value to this Cookie.
	 * 
	 * If you use a binary value, you may want to use BASE64 encoding.
	 * 
	 * With Version 0 cookies, values should not contain white space, brackets,
	 * parentheses, equals signs, commas, double quotes, slashes, question
	 * marks, at signs, colons, and semicolons. Empty values may not behave the
	 * same way on all browsers.
	 * 
	 * @param newValue the new value of the cookie
	 */
	public void setValue(String newValue) {
		this.value = newValue;
	}

	/**
	 * Gets the current value of this Cookie.
	 * 
	 * @return the current value of this Cookie
	 */
	public String getValue() {
		return value;
	}

	/**
	 * Returns the version of the protocol this cookie complies with. Version 1
	 * complies with RFC 2109, and version 0 complies with the original cookie
	 * specification drafted by Netscape. Cookies provided by a browser use and
	 * identify the browser's cookie version.
	 * 
	 * @return 0 if the cookie complies with the original Netscape
	 * specification; 1 if the cookie complies with RFC 2109
	 */
	public int getVersion() {
		return version;
	}

	/**
	 * 	Sets the version of the cookie protocol that this Cookie complies with.
	 * 
	 * Version 0 complies with the original Netscape cookie specification.
	 * Version 1 complies with RFC 2109.
	 * 
	 * Since RFC 2109 is still somewhat new, consider version 1 as experimental;
	 * do not use it yet on production sites.
	 * 
	 * @param v 0 if the cookie should comply with the original Netscape
	 * specification; 1 if the cookie should comply with RFC 2109.
	 */
	public void setVersion(int v) {
		this.version = v;
	}

	/**
	 * Overrides the standard java.lang.Object.clone method to return a copy
	 * of this Cookie.
	 */
	public Object clone() {
		Cookie returnValue = new Cookie(name, value);
		returnValue.comment = comment;
		returnValue.domain = domain;
		returnValue.maxAge = maxAge;
		returnValue.path = path;
		returnValue.secure = secure;
		returnValue.version = version;
		returnValue.httpOnly = httpOnly;
		
		return returnValue;
	}

	/**
	 * Marks or unmarks this Cookie as HttpOnly.
	 * 
	 * If isHttpOnly is set to true, this cookie is marked as HttpOnly, by
	 * adding the HttpOnly attribute to it.
	 * 
	 * HttpOnly cookies are not supposed to be exposed to client-side scripting
	 * code, and may therefore help mitigate certain kinds of cross-site
	 * scripting attacks.
	 * 
	 * @param isHttpOnly true if this cookie is to be marked as HttpOnly,
	 * false otherwise.
	 */
	public void setHttpOnly(boolean isHttpOnly) {
		this.httpOnly = isHttpOnly;
	}

	/**
	 * Checks whether this Cookie has been marked as HttpOnly.
	 * 
	 * @return true if this Cookie has been marked as HttpOnly, false otherwise.
	 */
	public boolean isHttpOnly() {
		return httpOnly;
	}
	
	/**
	 * Serialize the cookie so that it is ready for the response.
	 * 
	 * @return A String representation of the cookie so that it can be applied
	 * to the Set-Cookie header.
	 */
	public String serializeForResponse() {
		// TODO make sure we are not missing something for the Set-Cookie
		// parameters.
		StringBuilder sb = new StringBuilder();
		sb.append(name);
		sb.append("=");
		sb.append(value);
		if (path != null) {
			sb.append("; ");
			sb.append(HttpConstants.COOKIE_PATH_LABEL);
			sb.append("=");
			sb.append(path);
		}
		if (domain != null) {
			sb.append("; ");
			sb.append(HttpConstants.COOKIE_DOMAIN_LABEL);
			sb.append("=");
			sb.append(domain);
		}
		if (maxAge != -1) {
			sb.append("; expires ");
			Calendar expirationDate = Calendar.getInstance();
			expirationDate.setTime(new Date());
			expirationDate.add(Calendar.SECOND, maxAge);
			sb.append(DateHelper.getOldCookieFormat().format(
					expirationDate.getTime()));
		}
		return sb.toString();
	}

}
