package com.koutra.nodejava;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.util.Collection;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

/**
 * A HTTP request. Provides access to all the elements of the HTTP request, and
 * also provides an attribute map facility that maps Strings to Objects. The
 * attribute map is alive only during the duration of a request processing.
 * 
 * @author pftakas
 */
public interface HttpRequest {
	/**
	 * Returns the value of the named attribute as an Object, or null if no
	 * attribute of the given name exists.
	 * 
	 * Attributes can be set two ways. The container may set attributes to
	 * make available custom information about a request. For example, for
	 * requests made using HTTPS, the attribute request.X509Certificate can be
	 * used to retrieve information on the certificate of the client.
	 * Attributes can also be set programatically using
	 * HttpRequest#setAttribute. This allows information to be embedded into a
	 * request before delegating the request.
	 * 
	 * Attribute names should follow the same conventions as package names.
	 * 
	 * @param name a String specifying the name of the attribute
	 * @return an Object containing the value of the attribute, or null if the
	 * attribute does not exist
	 */
	Object getAttribute(String name);
	
	/**
	 * Returns a Set containing the names of the attributes available to this
	 * request. This method returns an empty Set if the request has no
	 * attributes available to it.
	 * 
	 * @return a Set of strings containing the names of the request's attributes
	 */
	Set<String> getAttributeNames();
	
	/**
	 * Returns the name of the character encoding used in the body of this
	 * request. This method returns null if the request does not specify a
	 * character encoding.
	 * 
	 * @return a String containing the name of the character encoding, or null
	 * if the request does not specify a character encoding
	 */
	String getCharacterEncoding();
	
	/**
	 * Overrides the name of the character encoding used in the body of this
	 * request. This method must be called prior to reading request parameters
	 * or reading input using getReader(). Otherwise, it has no effect.
	 *
	 * @param enc a String containing the name of the character encoding.
	 * @throws UnsupportedEncodingException if this ServletRequest is still in
	 * a state where a character encoding may be set, but the specified
	 * encoding is invalid
	 */
	void setCharacterEncoding(String enc) throws UnsupportedEncodingException;

	/**
	 * Returns the length, in bytes, of the request body and made available by
	 * the input stream, or -1 if the length is not known. Same as the value
	 * of the CGI variable CONTENT_LENGTH.
	 * 
	 * @return an integer containing the length of the request body or -1 if
	 * the length is not known
	 */
	int getContentLength();

	/**
	 * Returns the MIME type of the body of the request, or null if the type is
	 * not known. Same as the value of the CGI variable CONTENT_TYPE.
	 * 
	 * @return a String containing the name of the MIME type of the request, or
	 * null if the type is not known
	 */
	String getContentType();

	/**
	 * Retrieves the body of the request as binary data using an InputStream.
	 * Either this method or getReader() may be called to read the body, not
	 * both.
	 * 
	 * @return an InputStream object containing the body of the request
	 * @throws IllegalStateException - if the getReader() method has already 
	 * been called for this request
	 * @throws IOException if an input or output exception occurred
	 */
	InputStream getInputStream() throws IOException;

	/**
	 * Returns the value of a request parameter as a String, or null if the
	 * parameter does not exist. Request parameters are extra information sent
	 * with the request. Parameters are contained in the query string or
	 * posted form data.
	 * 
	 * You should only use this method when you are sure the parameter has only
	 * one value. If the parameter might have more than one value, use
	 * getParameterValues(String).
	 * 
	 * If you use this method with a multivalued parameter, the value returned
	 * is equal to the first value in the array returned by getParameterValues.
	 * 
	 * If the parameter data was sent in the request body, such as occurs with
	 * an HTTP POST request, then reading the body directly via getInputStream()
	 * or getReader() can interfere with the execution of this method.
	 * 
	 * @param name a String specifying the name of the parameter
	 * @return a String representing the single value of the parameter
	 */
	String getParameter(String name);

	/**
	 * Returns a Set of String objects containing the names of the parameters
	 * contained in this request. If the request has no parameters, the method
	 * returns an empty Set.
	 * 
	 * @return a Set of String objects, each String containing the name of a
	 * request parameter; or an empty Set if the request has no parameters
	 */
	Set<String> getParameterNames();

	/**
	 * Returns an array of String objects containing all of the values the
	 * given request parameter has, or null if the parameter does not exist.
	 * 
	 * If the parameter has a single value, the array has a length of 1.
	 * 
	 * @param name a String containing the name of the parameter whose value is
	 * requested
	 * @return an array of String objects containing the parameter's values
	 */
	String[] getParameterValues(String name);

	/**
	 * Returns a java.util.Map of the parameters of this request.
	 * 
	 * Request parameters are extra information sent with the request.
	 * Parameters are contained in the query string or posted form data.
	 * 
	 * @return an immutable Map containing parameter names as keys and
	 * parameter values as map values. The keys in the parameter map are of
	 * type String. The values in the parameter map are of type String array.
	 */
	Map<String,String[]> getParameterMap();

	/**
	 * Returns the name and version of the protocol the request uses in the
	 * form protocol/majorVersion.minorVersion, for example, HTTP/1.1. The
	 * value returned is the same as the value of the CGI variable
	 * SERVER_PROTOCOL.
	 * 
	 * @return a String containing the protocol name and version number
	 */
	String getProtocol();

	/**
	 * Returns the name of the scheme used to make this request, for example,
	 * http, https, or ftp. Different schemes have different rules for
	 * constructing URLs, as noted in RFC 1738.
	 * 
	 * @return a String containing the name of the scheme used to make this
	 * request
	 */
	String getScheme();

	/**
	 * Returns the host name of the server to which the request was sent. It
	 * is the value of the part before ":" in the Host header value, if any,
	 * or the resolved server name, or the server IP address.
	 * 
	 * @return a String containing the name of the server
	 */
	String getServerName();

	/**
	 * Returns the port number to which the request was sent. It is the value
	 * of the part after ":" in the Host header value, if any, or the server
	 * port where the client connection was accepted on.
	 * 
	 * @return an integer specifying the port number
	 */
	int getServerPort();

	/**
	 * Retrieves the body of the request as character data using a
	 * BufferedReader. The reader translates the character data according to
	 * the character encoding used on the body. Either this method or
	 * getInputStream() may be called to read the body, not both.
	 * 
	 * @return a BufferedReader containing the body of the request
	 * @throws UnsupportedEncodingException if the character set encoding used
	 * is not supported and the text cannot be decoded
	 * @throws IOException if an input or output exception occurred
	 */
	BufferedReader getReader() throws IOException;

	/**
	 * Returns the Internet Protocol (IP) address of the client or last proxy
	 * that sent the request. Same as the value of the CGI variable REMOTE_ADDR.
	 * 
	 * @return a String containing the IP address of the client that sent the request
	 */
	String getRemoteAddr();

	/**
	 * Returns the fully qualified name of the client or the last proxy that
	 * sent the request. If the engine cannot or chooses not to resolve the
	 * hostname (to improve performance), this method returns the dotted-string
	 * form of the IP address. Same as the value of the CGI variable REMOTE_HOST.
	 * 
	 * @return a String containing the fully qualified name of the client
	 */
	String getRemoteHost();

	/**
	 * Stores an attribute in this request. Attributes are reset between
	 * requests. This method is most often used in conjunction with delegation
	 * of the request.
	 * 
	 * Attribute names should follow the same conventions as package names.
	 * 
	 * If the object passed in is null, the effect is the same as calling
	 * removeAttribute(String).
	 * 
	 * @param name a String specifying the name of the attribute
	 * @param value the Object to be stored
	 */
	void setAttribute(String name, Object value);

	/**
	 * Removes an attribute from this request. This method is not generally
	 * needed as attributes only persist as long as the request is being handled.
	 * 
	 * Attribute names should follow the same conventions as package names.
	 *  
	 * @param name a String specifying the name of the attribute to remove
	 */
	void removeAttribute(String name);

	/**
	 * Returns the preferred Locale that the client will accept content in,
	 * based on the Accept-Language header. If the client request doesn't
	 * provide an Accept-Language header, this method returns the default
	 * locale for the server.
	 * 
	 * @return the preferred Locale for the client
	 */
	Locale getLocale();

	/**
	 * Returns a List of Locale objects indicating, in decreasing order
	 * starting with the preferred locale, the locales that are acceptable to
	 * the client based on the Accept-Language header. If the client request
	 * doesn't provide an Accept-Language header, this method returns an
	 * List containing one Locale, the default locale for the server.
	 * 
	 * @return
	 */
	List<Locale> getLocales();

	/**
	 * Returns a boolean indicating whether this request was made using a
	 * secure channel, such as HTTPS.
	 * 
	 * @return a boolean indicating if the request was made using a secure
	 * channel
	 */
	boolean isSecure();
	
	/**
	 * Returns an array containing all of the Cookie objects the client sent
	 * with this request. This method returns null if no cookies were sent.
	 * 
	 * @return an array of all the Cookies included with this request, or null
	 * if the request has no cookies
	 */
	Cookie[] getCookies();

	/**
	 * Returns the value of the specified request header as a long value that
	 * represents a Date object. Use this method with headers that contain
	 * dates, such as If-Modified-Since.
	 * 
	 * The date is returned as the number of milliseconds since January 1,
	 * 1970 GMT. The header name is case insensitive.
	 * 
	 * If the request did not have a header of the specified name, this method
	 * returns -1. If the header can't be converted to a date, the method
	 * throws an IllegalArgumentException.
	 * 
	 * @param name a String specifying the name of the header
	 * @return a long value representing the date specified in the header
	 * expressed as the number of milliseconds since January 1, 1970 GMT, or -1
	 * if the named header was not included with the request.
	 * @throws IllegalArgumentException If the header value can't be converted
	 * to a date.
	 */
	long getDateHeader(String name);

	/**
	 * Returns the value of the specified request header as a String. If the
	 * request did not include a header of the specified name, this method
	 * returns null. If there are multiple headers with the same name, this
	 * method returns the first head in the request. The header name is case
	 * insensitive. You can use this method with any request header.
	 * 
	 * @param name a String specifying the header name
	 * @return a String containing the value of the requested header, or null
	 * if the request does not have a header of that name
	 */
	String getHeader(String name);

	/**
	 * Returns all the values of the specified request header as a List
	 * of String objects.
	 * 
	 * Some headers, such as Accept-Language can be sent by clients as several
	 * headers each with a different value rather than sending the header as a
	 * comma separated list.
	 * 
	 * If the request did not include any headers of the specified name, this
	 * method returns an empty List. The header name is case insensitive.
	 * You can use this method with any request header.
	 * 
	 * @param name a String specifying the header name
	 * @return a list containing the values of the requested header.
	 * If the request does not have any headers of that name return an empty
	 * List. If the container does not allow access to header
	 * information, return null
	 */
	List<String> getHeaders(String name);

	/**
	 * Returns a set of all the header names this request contains. If
	 * the request has no headers, this method returns an empty set.
	 * 
	 * Some containers do not allow access to headers using
	 * this method, in which case this method returns null.
	 * 
	 * @return a set of all the header names sent with this request;
	 * if the request has no headers, an empty set; if the
	 * container does not allow clients to use this method, null.
	 */
	Set<String> getHeaderNames();

	/**
	 * Returns the value of the specified request header as an int. If the
	 * request does not have a header of the specified name, this method
	 * returns -1. If the header cannot be converted to an integer, this
	 * method throws a NumberFormatException.
	 * 
	 * The header name is case insensitive.
	 * 
	 * @param name a String specifying the name of a request header
	 * @return an integer expressing the value of the request header or -1 if
	 * the request doesn't have a header of this name.
	 * @throws NumberFormatException If the header value can't be converted
	 * to an int.
	 */
	int getIntHeader(String name);

	/**
	 * Returns the name of the HTTP method with which this request was made,
	 * for example, GET, POST, or PUT. Same as the value of the CGI variable
	 * REQUEST_METHOD.
	 * 
	 * @return a String specifying the name of the method with which this
	 * request was made
	 */
	String getMethod();

	/**
	 * Returns the part of the URL that is the path of the request (the part
	 * after the hostname:port part and before the query).
	 * 
	 * @return a String containing the path of the request.
	 */
	String getPath();

	/**
	 * Returns the query string that is contained in the request URL after the
	 * path. This method returns null if the URL does not have a query string.
	 * Same as the value of the CGI variable QUERY_STRING.
	 * 
	 * @return a String containing the query string or null if the URL contains
	 * no query string. The value is not decoded by the container.
	 */
	String getQueryString();

	/**
	 * Returns the part of this request's URL from the protocol name up to the
	 * query string in the first line of the HTTP request. The web container
	 * does not decode this String. For example:
	 * 
	 * <table>
	 *     <tr>
	 *         <td>First line of HTTP request</td>
	 *         <td>Returned Value</td>
	 *     </tr>
	 *     <tr>
	 *         <td>POST /some/path.html HTTP/1.1</td>
	 *         <td>/some/path.html</td>
	 *     </tr>
	 *     <tr>
	 *         <td>GET http://foo.bar/a.html HTTP/1.0</td>
	 *         <td>/a.html</td>
	 *     </tr>
	 *     <tr>
	 *         <td>HEAD /xyz?a=b HTTP/1.1</td>
	 *         <td>/xyz</td>
	 *     </tr>
	 * </table>
	 * 
	 * To reconstruct an URL with a scheme and host, use HttpUtils#getRequestURL.
	 * 
	 * @return a String containing the part of the URL from the protocol name
	 * up to the query string
	 */
	String getRequestURI();

	/**
	 * Reconstructs the URL the client used to make the request. The returned
	 * URL contains a protocol, server name, port number, and server path, but
	 * it does not include query string parameters.
	 * 
	 * Because this method returns a StringBuffer, not a string, you can modify
	 * the URL easily, for example, to append query parameters.
	 * 
	 * This method is useful for creating redirect messages and for reporting
	 * errors.
	 * 
	 * @return a StringBuffer object containing the reconstructed URL
	 */
	StringBuffer getRequestURL();

	/**
	 * Returns the current HttpSession associated with this request or, if
	 * there is no current session and create is true, returns a new session.
	 * 
	 * If create is false and the request has no valid HttpSession, this method
	 * returns null.
	 * 
	 * To make sure the session is properly maintained, you must call this
	 * method before the response is committed. If the container is using
	 * cookies to maintain session integrity and is asked to create a new
	 * session when the response is committed, an IllegalStateException is
	 * thrown.
	 * 
	 * @param create true to create a new session for this request if necessary;
	 * false to return null if there's no current session
	 * @return the HttpSession associated with this request or null if create
	 * is false and the request has no valid session
	 */
	HttpSession getSession(boolean create);

	/**
	 * Returns the current session associated with this request, or if the
	 * request does not have a session, creates one.
	 * 
	 * @return the HttpSession associated with this request
	 */
	HttpSession getSession();

	/**
	 * Gets all the Part components of this request, provided that it is of
	 * type multipart/form-data.
	 * 
	 * If this request is of type multipart/form-data, but does not contain
	 * any Part components, the returned Collection will be empty.
	 * 
	 * Any changes to the returned Collection must not affect this HttpRequest.
	 * 
	 * @return
	 * @throws IOException if an I/O error occurred during the retrieval of the
	 * Part components of this request
	 * @throws NodeJavaException if this request is not of type
	 * multipart/form-data
	 * @throws IllegalStateException if the request body is larger than
	 * maxRequestSize, or any Part in the request is larger than maxFileSize
	 */
	Collection<Part> getParts() throws IOException, NodeJavaException;

	/**
	 * Gets the Part with the given name.
	 * 
	 * @param name the name of the requested Part
	 * @return The Part with the given name, or null if this request is of type
	 * multipart/form-data, but does not contain the requested Part.
	 * @throws IOException if an I/O error occurred during the retrieval of the
	 * requested Part
	 * @throws NodeJavaException if this request is not of type
	 * multipart/form-data
	 * @throws IllegalStateException if the request body is larger than
	 * maxRequestSize, or any Part in the request is larger than maxFileSize
	 */
	Part getPart(String name) throws IOException, NodeJavaException;
}
