package com.koutra.nodejava;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Locale;

/**
 * A HTTP response. Provides access to the various HTTP response elements,
 * provides constants for the various HTTP status codes, as well as helper
 * methods sending error responses and redirects.
 * 
 * @author pftakas
 */
public interface HttpResponse {
    /*
     * Server status codes; see RFC 2068.
     */

    /**
     * Status code (100) indicating the client can continue.
     */

    public static final int SC_CONTINUE = 100;

    
    /**
     * Status code (101) indicating the server is switching protocols
     * according to Upgrade header.
     */

    public static final int SC_SWITCHING_PROTOCOLS = 101;

    /**
     * Status code (200) indicating the request succeeded normally.
     */

    public static final int SC_OK = 200;

    /**
     * Status code (201) indicating the request succeeded and created
     * a new resource on the server.
     */

    public static final int SC_CREATED = 201;

    /**
     * Status code (202) indicating that a request was accepted for
     * processing, but was not completed.
     */

    public static final int SC_ACCEPTED = 202;

    /**
     * Status code (203) indicating that the meta information presented
     * by the client did not originate from the server.
     */

    public static final int SC_NON_AUTHORITATIVE_INFORMATION = 203;

    /**
     * Status code (204) indicating that the request succeeded but that
     * there was no new information to return.
     */

    public static final int SC_NO_CONTENT = 204;

    /**
     * Status code (205) indicating that the agent <em>SHOULD</em> reset
     * the document view which caused the request to be sent.
     */

    public static final int SC_RESET_CONTENT = 205;

    /**
     * Status code (206) indicating that the server has fulfilled
     * the partial GET request for the resource.
     */

    public static final int SC_PARTIAL_CONTENT = 206;

    /**
     * Status code (300) indicating that the requested resource
     * corresponds to any one of a set of representations, each with
     * its own specific location.
     */

    public static final int SC_MULTIPLE_CHOICES = 300;

    /**
     * Status code (301) indicating that the resource has permanently
     * moved to a new location, and that future references should use a
     * new URI with their requests.
     */

    public static final int SC_MOVED_PERMANENTLY = 301;

    /**
     * Status code (302) indicating that the resource has temporarily
     * moved to another location, but that future references should
     * still use the original URI to access the resource.
     *
     * This definition is being retained for backwards compatibility.
     * SC_FOUND is now the preferred definition.
     */

    public static final int SC_MOVED_TEMPORARILY = 302;

    /**
    * Status code (302) indicating that the resource reside
    * temporarily under a different URI. Since the redirection might
    * be altered on occasion, the client should continue to use the
    * Request-URI for future requests.(HTTP/1.1) To represent the
    * status code (302), it is recommended to use this variable.
    */

    public static final int SC_FOUND = 302;

    /**
     * Status code (303) indicating that the response to the request
     * can be found under a different URI.
     */

    public static final int SC_SEE_OTHER = 303;

    /**
     * Status code (304) indicating that a conditional GET operation
     * found that the resource was available and not modified.
     */

    public static final int SC_NOT_MODIFIED = 304;

    /**
     * Status code (305) indicating that the requested resource
     * <em>MUST</em> be accessed through the proxy given by the
     * <code><em>Location</em></code> field.
     */

    public static final int SC_USE_PROXY = 305;

     /**
     * Status code (307) indicating that the requested resource 
     * resides temporarily under a different URI. The temporary URI
     * <em>SHOULD</em> be given by the <code><em>Location</em></code> 
     * field in the response.
     */

     public static final int SC_TEMPORARY_REDIRECT = 307;

    /**
     * Status code (400) indicating the request sent by the client was
     * syntactically incorrect.
     */

    public static final int SC_BAD_REQUEST = 400;

    /**
     * Status code (401) indicating that the request requires HTTP
     * authentication.
     */

    public static final int SC_UNAUTHORIZED = 401;

    /**
     * Status code (402) reserved for future use.
     */

    public static final int SC_PAYMENT_REQUIRED = 402;

    /**
     * Status code (403) indicating the server understood the request
     * but refused to fulfill it.
     */

    public static final int SC_FORBIDDEN = 403;

    /**
     * Status code (404) indicating that the requested resource is not
     * available.
     */

    public static final int SC_NOT_FOUND = 404;

    /**
     * Status code (405) indicating that the method specified in the
     * <code><em>Request-Line</em></code> is not allowed for the resource
     * identified by the <code><em>Request-URI</em></code>.
     */

    public static final int SC_METHOD_NOT_ALLOWED = 405;

    /**
     * Status code (406) indicating that the resource identified by the
     * request is only capable of generating response entities which have
     * content characteristics not acceptable according to the accept
     * headers sent in the request.
     */

    public static final int SC_NOT_ACCEPTABLE = 406;

    /**
     * Status code (407) indicating that the client <em>MUST</em> first
     * authenticate itself with the proxy.
     */

    public static final int SC_PROXY_AUTHENTICATION_REQUIRED = 407;

    /**
     * Status code (408) indicating that the client did not produce a
     * request within the time that the server was prepared to wait.
     */

    public static final int SC_REQUEST_TIMEOUT = 408;

    /**
     * Status code (409) indicating that the request could not be
     * completed due to a conflict with the current state of the
     * resource.
     */

    public static final int SC_CONFLICT = 409;

    /**
     * Status code (410) indicating that the resource is no longer
     * available at the server and no forwarding address is known.
     * This condition <em>SHOULD</em> be considered permanent.
     */

    public static final int SC_GONE = 410;

    /**
     * Status code (411) indicating that the request cannot be handled
     * without a defined <code><em>Content-Length</em></code>.
     */

    public static final int SC_LENGTH_REQUIRED = 411;

    /**
     * Status code (412) indicating that the precondition given in one
     * or more of the request-header fields evaluated to false when it
     * was tested on the server.
     */

    public static final int SC_PRECONDITION_FAILED = 412;

    /**
     * Status code (413) indicating that the server is refusing to process
     * the request because the request entity is larger than the server is
     * willing or able to process.
     */

    public static final int SC_REQUEST_ENTITY_TOO_LARGE = 413;

    /**
     * Status code (414) indicating that the server is refusing to service
     * the request because the <code><em>Request-URI</em></code> is longer
     * than the server is willing to interpret.
     */

    public static final int SC_REQUEST_URI_TOO_LONG = 414;

    /**
     * Status code (415) indicating that the server is refusing to service
     * the request because the entity of the request is in a format not
     * supported by the requested resource for the requested method.
     */

    public static final int SC_UNSUPPORTED_MEDIA_TYPE = 415;

    /**
     * Status code (416) indicating that the server cannot serve the
     * requested byte range.
     */

    public static final int SC_REQUESTED_RANGE_NOT_SATISFIABLE = 416;

    /**
     * Status code (417) indicating that the server could not meet the
     * expectation given in the Expect request header.
     */

    public static final int SC_EXPECTATION_FAILED = 417;

    /**
     * Status code (500) indicating an error inside the HTTP server
     * which prevented it from fulfilling the request.
     */

    public static final int SC_INTERNAL_SERVER_ERROR = 500;

    /**
     * Status code (501) indicating the HTTP server does not support
     * the functionality needed to fulfill the request.
     */

    public static final int SC_NOT_IMPLEMENTED = 501;

    /**
     * Status code (502) indicating that the HTTP server received an
     * invalid response from a server it consulted when acting as a
     * proxy or gateway.
     */

    public static final int SC_BAD_GATEWAY = 502;

    /**
     * Status code (503) indicating that the HTTP server is
     * temporarily overloaded, and unable to handle the request.
     */

    public static final int SC_SERVICE_UNAVAILABLE = 503;

    /**
     * Status code (504) indicating that the server did not receive
     * a timely response from the upstream server while acting as
     * a gateway or proxy.
     */

    public static final int SC_GATEWAY_TIMEOUT = 504;

    /**
     * Status code (505) indicating that the server does not support
     * or refuses to support the HTTP protocol version that was used
     * in the request message.
     */

    public static final int SC_HTTP_VERSION_NOT_SUPPORTED = 505;

	/**
	 * Returns the name of the character encoding (MIME charset) used for the
	 * body sent in this response. The character encoding may have been
	 * specified explicitly using the setCharacterEncoding(java.lang.String)
	 * or setContentType(java.lang.String) methods, or implicitly using the
	 * setLocale(java.util.Locale) method. Explicit specifications take
	 * precedence over implicit specifications. Calls made to these methods
	 * after getWriter has been called or after the response has been committed
	 * have no effect on the character encoding. If no character encoding has
	 * been specified, ISO-8859-1 is returned.
	 * 
	 * See RFC 2047 (http://www.ietf.org/rfc/rfc2047.txt) for more information
	 * about character encoding and MIME.
	 * 
	 * @returna String specifying the name of the character encoding, for
	 * example, UTF-8.
	 */
	String getCharacterEncoding();

	/**
	 * Returns the content type used for the MIME body sent in this response.
	 * The content type proper must have been specified using
	 * setContentType(java.lang.String) before the response is committed. If
	 * no content type has been specified, this method returns null. If a
	 * content type has been specified, and a character encoding has been
	 * explicitly or implicitly specified as described in getCharacterEncoding()
	 * or getWriter() has been called, the charset parameter is included in the
	 * string returned. If no character encoding has been specified, the
	 * charset parameter is omitted.
	 * 
	 * @return a String specifying the content type, for example, text/html;
	 * charset=UTF-8, or null.
	 */
	String getContentType();

	/**
	 * This method sets the flag that informs the server whether to use
	 * chunked encoding or not. Normally this is set on a server wide level.
	 * However, this is provided for overriding the server wide setting on a
	 * per request basis.
	 * 
	 * @param chunkedEncoding true if the response is to use chunked encoding,
	 * false otherwise.
	 */
	void setChunkedEncoding(boolean chunkedEncoding);
	
	/**
	 * Returns a ServletOutputStream suitable for writing binary data in the
	 * response. The container does not encode the binary data.
	 * 
	 * Calling flush() on the ServletOutputStream commits the response. Either
	 * this method or getWriter() may be called to write the body, not both.
	 * 
	 * @return a ServletOutputStream for writing binary data
	 * @throws IOException if an input or output exception occurred
	 * @throws IllegalStateException if the getWriter method has been called
	 * on this response
	 */
	OutputStream getOutputStream() throws IOException;

	/**
	 * Returns a PrintWriter object that can send character text to the client.
	 * The PrintWriter uses the character encoding returned by
	 * getCharacterEncoding(). If the response's character encoding has not
	 * been specified as described in getCharacterEncoding (i.e., the method
	 * just returns the default value ISO-8859-1), getWriter updates it to
	 * ISO-8859-1.
	 * 
	 * Calling flush() on the PrintWriter commits the response.
	 * 
	 * Either this method or getOutputStream() may be called to write the body,
	 * not both.
	 * 
	 * @return a PrintWriter object that can return character data to the client
	 * @throws IOException if an input or output exception occurred
	 * @throws UnsupportedEncodingException if the character encoding returned
	 * by getCharacterEncoding cannot be used
	 * @throws IllegalStateException if the getOutputStream method has already
	 * been called for this response object
	 */
	PrintWriter getWriter() throws IOException;

	/**
	 * Sets the character encoding (MIME charset) of the response being sent
	 * to the client, for example, to UTF-8. If the character encoding has
	 * already been set by setContentType(java.lang.String) or
	 * setLocale(java.util.Locale), this method overrides it. Calling
	 * setContentType(java.lang.String) with the String of text/html and
	 * calling this method with the String of UTF-8 is equivalent with calling
	 * setContentType with the String of text/html; charset=UTF-8.
	 * 
	 * This method can be called repeatedly to change the character encoding.
	 * This method has no effect if it is called after getWriter has been
	 * called or after the response has been committed.
	 * 
	 * Containers must communicate the character encoding used for the
	 * response's writer to the client if the protocol provides a way for doing
	 * so. In the case of HTTP, the character encoding is communicated as part
	 * of the Content-Type header for text media types. Note that the character
	 * encoding cannot be communicated via HTTP headers if the content type is
	 * not specified; however, it is still used to encode text written via the
	 * response's writer.
	 * 
	 * @param charset a String specifying only the character set defined by IANA
	 * Character Sets (http://www.iana.org/assignments/character-sets)
	 */
	void setCharacterEncoding(java.lang.String charset);

	/**
	 * Sets the length of the content body in the response This method sets the
	 * HTTP Content-Length header.
	 * 
	 * @param len an integer specifying the length of the content being
	 * returned to the client; sets the Content-Length header
	 */
	void setContentLength(int len);

	/**
	 * Sets the content type of the response being sent to the client, if the
	 * response has not been committed yet. The given content type may include
	 * a character encoding specification, for example, text/html;charset=UTF-8.
	 * The response's character encoding is only set from the given content type
	 * if this method is called before getWriter is called.
	 * 
	 * This method may be called repeatedly to change content type and character
	 * encoding. This method has no effect if called after the response has been
	 * committed. It does not set the response's character encoding if it is
	 * called after getWriter has been called or after the response has been
	 * committed.
	 * 
	 * Containers must communicate the content type and the character encoding
	 * used for the response's writer to the client if the protocol provides a
	 * way for doing so. In the case of HTTP, the Content-Type header is used.
	 * 
	 * @param type a String specifying the MIME type of the content
	 */
	void setContentType(String type);

	/**
	 * Sets the preferred buffer size for the body of the response. The
	 * container will use a buffer at least as large as the size requested.
	 * The actual buffer size used can be found using getBufferSize.
	 * 
	 * A larger buffer allows more content to be written before anything is
	 * actually sent, thus providing the server with more time to set
	 * appropriate status codes and headers. A smaller buffer decreases server
	 * memory load and allows the client to start receiving data more quickly.
	 * 
	 * This method must be called before any response body content is written;
	 * if content has been written or the response object has been committed,
	 * this method throws an IllegalStateException.
	 * 
	 * @param size the preferred buffer size
	 * @throws IllegalStateException if this method is called after content has
	 * been written
	 */
	void setBufferSize(int size);

	/**
	 * Returns the actual buffer size used for the response. If no buffering
	 * is used, this method returns 0.
	 * 
	 * @return the actual buffer size used
	 */
	int getBufferSize();

	/**
	 * Forces any content in the buffer to be written to the client. A call
	 * to this method automatically commits the response, meaning the status
	 * code and headers will be written.
	 * 
	 * @throws IOException
	 */
	void flushBuffer() throws IOException;
	
	/**
	 * This method is equivalent to flushBuffer(), but it additionally signals
	 * that the end of the response has been reached. This is only useful when
	 * using chunked encoding.
	 * 
	 * @throws IOException
	 */
	void terminate() throws IOException;

	/**
	 * Clears the content of the underlying buffer in the response without
	 * clearing headers or status code. If the response has been committed,
	 * this method throws an IllegalStateException.
	 */
	void resetBuffer();

	/**
	 * Returns a boolean indicating if the response has been committed. A
	 * committed response has already had its status code and headers written.
	 * 
	 * @return a boolean indicating if the response has been committed
	 */
	boolean isCommitted();

	/**
	 * Clears any data that exists in the buffer as well as the status code
	 * and headers. If the response has been committed, this method throws an
	 * IllegalStateException.
	 */
	void reset();

	/**
	 * Sets the locale of the response, if the response has not been committed
	 * yet. It also sets the response's character encoding appropriately for
	 * the locale, if the character encoding has not been explicitly set using
	 * setContentType(java.lang.String) or setCharacterEncoding(java.lang.String),
	 * getWriter hasn't been called yet, and the response hasn't been committed
	 * yet. If the deployment descriptor contains a locale-encoding-mapping-list
	 * element, and that element provides a mapping for the given locale, that
	 * mapping is used. Otherwise, the mapping from locale to character
	 * encoding is container dependent.
	 * 
	 * This method may be called repeatedly to change locale and character
	 * encoding. The method has no effect if called after the response has been
	 * committed. It does not set the response's character encoding if it is
	 * called after setContentType(java.lang.String) has been called with a
	 * charset specification, after setCharacterEncoding(java.lang.String) has
	 * been called, after getWriter has been called, or after the response has
	 * been committed.
	 * 
	 * Containers must communicate the locale and the character encoding used
	 * for the response's writer to the client if the protocol provides a way
	 * for doing so. In the case of HTTP, the locale is communicated via the
	 * Content-Language header, the character encoding as part of the
	 * Content-Type header for text media types. Note that the character
	 * encoding cannot be communicated via HTTP headers if the server does not
	 * specify a content type; however, it is still used to encode text written
	 * via the response's writer.
	 * 
	 * @param loc
	 */
	void setLocale(Locale loc);

	/**
	 * Returns the locale specified for this response using the
	 * setLocale(java.util.Locale) method. Calls made to setLocale after the
	 * response is committed have no effect. If no locale has been specified,
	 * the container's default locale is returned.
	 * @return
	 */
	Locale getLocale();

	   
    /**
     * Adds the specified cookie to the response.  This method can be called
     * multiple times to set more than one cookie.
     *
     * @param cookie the Cookie to return to the client
     */
    public void addCookie(Cookie cookie);

    /**
     * Returns a boolean indicating whether the named response header 
     * has already been set.
     * 
     * @param name the header name
     * @return <code>true</code> if the named response header has already
     * been set; <code>false</code> otherwise
     */
    public boolean containsHeader(String name);

    /**
     * Encodes the specified URL by including the session ID in it,
     * or, if encoding is not needed, returns the URL unchanged.
     * The implementation of this method includes the logic to
     * determine whether the session ID needs to be encoded in the URL.
     * For example, if the browser supports cookies, or session
     * tracking is turned off, URL encoding is unnecessary.
     * 
     * <p>For robust session tracking, all URLs emitted by a processor 
     * should be run through this
     * method.  Otherwise, URL rewriting cannot be used with browsers
     * which do not support cookies.
     *
     * @param url the url to be encoded.
     * @return the encoded URL if encoding is needed; the unchanged
     * URL otherwise.
     */
    public String encodeURL(String url);

    /**
     * Encodes the specified URL for use in the
     * <code>sendRedirect</code> method or, if encoding is not needed,
     * returns the URL unchanged.  The implementation of this method
     * includes the logic to determine whether the session ID
     * needs to be encoded in the URL.  Because the rules for making
     * this determination can differ from those used to decide whether to
     * encode a normal link, this method is separated from the
     * <code>encodeURL</code> method.
     * 
     * <p>All URLs sent to the <code>HttpServletResponse.sendRedirect</code>
     * method should be run through this method.  Otherwise, URL
     * rewriting cannot be used with browsers which do not support
     * cookies.
     *
     * @param url the url to be encoded.
     * @return the encoded URL if encoding is needed; the unchanged
     * URL otherwise.
     */
    public String encodeRedirectURL(String url);

    /**
     * Sends an error response to the client using the specified
     * status.  The server defaults to creating the
     * response to look like an HTML-formatted server error page
     * containing the specified message, setting the content type
     * to "text/html", leaving cookies and other headers unmodified.
     *
     * If an error-page declaration has been made for the web application
     * corresponding to the status code passed in, it will be served back in 
     * preference to the suggested msg parameter. 
     *
     * <p>If the response has already been committed, this method throws 
     * an IllegalStateException.
     * After using this method, the response should be considered
     * to be committed and should not be written to.
     *
     * @param sc the error status code
     * @param msg the descriptive message
     * @exception IOException If an input or output exception occurs
     * @exception IllegalStateException If the response was committed
     */
    public void sendError(int sc, String msg) throws IOException;

    /**
     * Sends an error response to the client using the specified status
     * code and clearing the buffer. 
     * <p>If the response has already been committed, this method throws 
     * an IllegalStateException.
     * After using this method, the response should be considered
     * to be committed and should not be written to.
     *
     * @param sc the error status code
     * @exception IOException If an input or output exception occurs
     * @exception IllegalStateException If the response was committed before
     * this method call
     */
    public void sendError(int sc) throws IOException;

    /**
     * Sends a temporary redirect response to the client using the
     * specified redirect location URL.  This method can accept relative URLs;
     * the container must convert the relative URL to an absolute URL
     * before sending the response to the client. If the location is relative 
     * without a leading '/' the container interprets it as relative to
     * the current request URI. If the location is relative with a leading
     * '/' the container interprets it as relative to the container root.
     *
     * <p>If the response has already been committed, this method throws 
     * an IllegalStateException.
     * After using this method, the response should be considered
     * to be committed and should not be written to.
     *
     * @param location the redirect location URL
     * @exception IOException If an input or output exception occurs
     * @exception IllegalStateException If the response was committed or
     * if a partial URL is given and cannot be converted into a valid URL
     */
    public void sendRedirect(String location) throws IOException;
    
    /**
     * Sets a response header with the given name and
     * date-value.  The date is specified in terms of
     * milliseconds since the epoch.  If the header had already
     * been set, the new value overwrites the previous one.  The
     * <code>containsHeader</code> method can be used to test for the
     * presence of a header before setting its value.
     * 
     * @param name the name of the header to set
     * @param date the assigned date value
     */
    public void setDateHeader(String name, long date);
    
    /**
     * Adds a response header with the given name and
     * date-value.  The date is specified in terms of
     * milliseconds since the epoch.  This method allows response headers 
     * to have multiple values.
     * 
     * @param name the name of the header to set
     * @param date the additional date value
     */
    public void addDateHeader(String name, long date);
    
    /**
     * Sets a response header with the given name and value.
     * If the header had already been set, the new value overwrites the
     * previous one.  The <code>containsHeader</code> method can be
     * used to test for the presence of a header before setting its
     * value.
     * 
     * @param name the name of the header
     * @param value the header value  If it contains octet string,
     *		it should be encoded according to RFC 2047
     *		(http://www.ietf.org/rfc/rfc2047.txt)
     */
    public void setHeader(String name, String value);
    
    /**
     * Adds a response header with the given name and value.
     * This method allows response headers to have multiple values.
     * 
     * @param name the name of the header
     * @param value the additional header value   If it contains
     *		octet string, it should be encoded
     *		according to RFC 2047
     *		(http://www.ietf.org/rfc/rfc2047.txt)
     */
    public void addHeader(String name, String value);

    /**
     * Sets a response header with the given name and
     * integer value.  If the header had already been set, the new value
     * overwrites the previous one.  The <code>containsHeader</code>
     * method can be used to test for the presence of a header before
     * setting its value.
     *
     * @param name the name of the header
     * @param value the assigned integer value
     */
    public void setIntHeader(String name, int value);

    /**
     * Adds a response header with the given name and
     * integer value.  This method allows response headers to have multiple
     * values.
     *
     * @param name the name of the header
     * @param value the assigned integer value
     */
    public void addIntHeader(String name, int value);


    
    /**
     * Sets the status code for this response.  This method is used to
     * set the return status code when there is no error (for example,
     * for the status codes SC_OK or SC_MOVED_TEMPORARILY).  If there
     * is an error, and the caller wishes to invoke an error-page defined
     * in the web application, the <code>sendError</code> method should be used
     * instead.
     * <p> The container clears the buffer and sets the Location header, preserving
     * cookies and other headers.
     *
     * @param sc the status code
     */
    public void setStatus(int sc);
  
}
