package com.koutra.nodejava;

import java.util.Set;

/**
 * Represents an HTTP session (the server side object that is associated with
 * the session). A session is created upon first request (usually by code that
 * handles authentication, but this is up to the application). It is renewed
 * upon each subsequent request from the client and it is terminated either by
 * an explicit call to the HttpSession#invalidate method, or after a certain
 * amount of time has passed with no requests coming in from the client
 * (specified by the HttpSession#setMaxInactiveInterval method). Also this
 * class provides an attribute map facility for the application to store any
 * session related state. Care should be taken to keep the session related
 * state to a minimum for performance and scalability.
 * 
 * @author pftakas
 */
public interface HttpSession {

	/**
	 * Returns the object bound with the specified name in this session, or
	 * null if no object is bound under the name.
	 * 
	 * @param name a string specifying the name of the object
	 * @return the object with the specified name
	 * @throws IllegalStateException if this method is called on an invalidated
	 * session
	 */
	Object getAttribute(String name);

	/**
	 * Returns a Set of String objects containing the names of all the objects
	 * bound to this session.
	 * 
	 * @return a Set of String objects specifying the names of all the objects
	 * bound to this session
	 * @throws IllegalStateException if this method is called on an invalidated
	 * session
	 */
	Set<String> getAttributeNames();
	
	/**
	 * Binds an object to this session, using the name specified. If an object
	 * of the same name is already bound to the session, the object is replaced.
	 * 
	 * If the value passed in is null, this has the same effect as calling
	 * removeAttribute().
	 * 
	 * @param name the name to which the object is bound; cannot be null
	 * @param value the object to be bound
	 * @throws IllegalStateException if this method is called on an invalidated
	 * session
	 */
	void setAttribute(String name, Object value);

	/**
	 * Removes the object bound with the specified name from this session. If
	 * the session does not have an object bound with the specified name, this
	 * method does nothing.
	 * 
	 * @param name the name of the object to remove from this session
	 * @throws IllegalStateException if this method is called on an invalidated
	 * session
	 */
	void removeAttribute(String name);

	/**
	 * Returns the time when this session was created, measured in milliseconds
	 * since midnight January 1, 1970 GMT.
	 * 
	 * @return a long specifying when this session was created, expressed in
	 * milliseconds since 1/1/1970 GMT.
	 * @throws IllegalStateException if this method is called on an invalidated
	 * session.
	 */
	long getCreationTime();

	/**
	 * Returns a string containing the unique identifier assigned to this
	 * session. The identifier is assigned by the container and is
	 * implementation dependent.
	 * 
	 * @return a string specifying the identifier assigned to this session.
	 */
	String getId();

	/**
	 * Returns the last time the client sent a request associated with this
	 * session, as the number of milliseconds since midnight January 1, 1970
	 * GMT, and marked by the time the container received the request.
	 * 
	 * Actions that your application takes, such as getting or setting a value
	 * associated with the session, do not affect the access time.
	 * 
	 * @return a long representing the last time the client sent a request
	 * associated with this session, expressed in milliseconds since 1/1/1970
	 * GMT.
	 * @throws IllegalStateException if this method is called on an invalidated
	 * session.
	 */
	long getLastAccessedTime();

	/**
	 * Specifies the time, in seconds, between client requests before the 
	 * container will invalidate this session.
	 * 
	 * An interval value of zero or less indicates that the session should
	 * never timeout.
	 * 
	 * @param interval an integer specifying the number of seconds.
	 */
	void setMaxInactiveInterval(int interval);

	/**
	 * Returns the maximum time interval, in seconds, that the container will
	 * keep this session open between client accesses. After this interval,
	 * the container will invalidate the session. The maximum time interval
	 * can be set with the setMaxInactiveInterval method.
	 * 
	 * A return value of zero or less indicates that the session will never timeout.
	 * 
	 * @return an integer specifying the number of seconds this session
	 * remains open between client requests
	 */
	int getMaxInactiveInterval();
	
	/**
	 * Checks to see if the session is still valid. A session is valid if the
	 * last access time has happened less than the maximum time interval ago.
	 * 
	 * @return true if the session is still valid, false otherwise.
	 */
	boolean isValid();

	/**
	 * Invalidates this session then unbinds any objects bound to it.
	 * 
	 * @throws IllegalStateException if this method is called on an invalidated
	 * session.
	 */
	void invalidate();

	/**
	 * Returns true if the client does not yet know about the session or if
	 * the client chooses not to join the session. For example, if the server
	 * used only cookie-based sessions, and the client had disabled the use of
	 * cookies, then a session would be new on each request.
	 * 
	 * @return true if the server has created a session, but the client has
	 * not yet joined.
	 * @throws IllegalStateException if this method is called on an invalidated
	 * session.
	 */
	boolean isNew();
}
