package com.koutra.nodejava;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import javax.net.ServerSocketFactory;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.pool.impl.GenericObjectPool;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.SAXException;

import com.koutra.nodejava.impl.SessionContainer;
import com.koutra.nodejava.registry.PatternRegistry;
import com.koutra.nodejava.registry.Transport;
import com.koutra.nodejava.utils.ProcessClientObjectFactory;
import com.koutra.nodejava.utils.ProcessClientTransportObjectFactory;

/**
 * The main class of the NodeJava project. Starts the HTTP server and listens
 * for incoming requests.
 * 
 * @author pftakas
 */
public class NodeJava implements Runnable {
	private static final Logger logger = LoggerFactory.getLogger(NodeJava.class);
	
	protected PatternRegistry patternRegistry;
	protected SessionContainer sessionContainer;
	protected Map<String, NodeJavaFacility> facilitiesMap;
	protected ExecutorService processClientExecutorService;
	protected GenericObjectPool<ProcessClient> processClientPool;
	protected GenericObjectPool<ProcessClientTransportAgnostic> processClientTransportPool;
	
	/**
	 * Initializing constructor.
	 * @throws IllegalAccessException 
	 * @throws InstantiationException 
	 * @throws ClassNotFoundException 
	 * @throws IOException 
	 * @throws SAXException 
	 * @throws ParserConfigurationException 
	 * @throws NodeJavaException 
	 */
	public NodeJava(String externConfigFile) throws ParserConfigurationException, SAXException,
			IOException, ClassNotFoundException, InstantiationException,
			IllegalAccessException, NodeJavaException {
		facilitiesMap = new HashMap<String, NodeJavaFacility>();
		patternRegistry = new PatternRegistry(externConfigFile, this);
		sessionContainer = new SessionContainer();
		processClientExecutorService = Executors.newCachedThreadPool();
		processClientPool = new GenericObjectPool<ProcessClient>(
				new ProcessClientObjectFactory(),
				patternRegistry.getTransport().getProcessClientPoolSize());
		processClientTransportPool = new GenericObjectPool<ProcessClientTransportAgnostic>(
				new ProcessClientTransportObjectFactory(),
				patternRegistry.getTransport().getProcessClientTransportPoolSize());
	}
	
	/**
	 * Getter method for the executor service of process clients.
	 * @return the process client executor service to be used by the process
	 * clients to run themselves on separate threads.
	 */
	public ExecutorService getProcessClientExecutorService() {
		return processClientExecutorService;
	}
	
	/**
	 * Helper method to release a process client back to the pool.
	 * @param processClient the process client to release.
	 * @throws Exception 
	 */
	public void releaseProcessClient(ProcessClient processClient)
			throws Exception {
		processClientPool.returnObject(processClient);
	}
	
	/**
	 * Helper method to acquire a process client transport.
	 * @return
	 * @throws InterruptedException
	 */
	public ProcessClientTransportAgnostic acquireProcessClientTransport()
			throws Exception {
		return processClientTransportPool.borrowObject();
	}
	
	/**
	 * Helper method to release a process client transport.
	 * @param processClientTransport
	 * @throws InterruptedException
	 */
	public void releaseProcessClientTransport(
			ProcessClientTransportAgnostic processClientTransport)
					throws Exception {
		processClientTransportPool.returnObject(processClientTransport);
	}
	
	/**
	 * Accessor for the pattern registry associated with this instance.
	 * @return
	 */
	public PatternRegistry getPatternRegistry() {
		return patternRegistry;
	}
	
	/**
	 * Accessor for the session container associated with this instance.
	 * @return
	 */
	public SessionContainer getSessionContainer() {
		return sessionContainer;
	}
	
	/**
	 * Registers a facility with the NodeJava node.
	 * @param facilityName the facility name.
	 * @param facility the facility instance.
	 * @throws NodeJavaException iff a facility with the same name has already
	 * been registered with this NodeJava node.
	 */
	public void registerFacility(String facilityName, NodeJavaFacility facility) throws NodeJavaException {
		if (facilitiesMap.containsKey(facilityName)) {
			throw new NodeJavaException("Attempt to register a facility that " +
					"is already registered: " + facilityName);
		}
		facilitiesMap.put(facilityName, facility);
	}
	
	/**
	 * Accessor for a facility with the given name.
	 * @param facilityName the facility name.
	 * @return the facility mapped to the facility name, or null if no such 
	 * facility exists.
	 */
	public NodeJavaFacility retrieveFacility(String facilityName) {
		return facilitiesMap.get(facilityName);
	}

	/**
	 * Create a simple server socket.
	 * 
	 * @return a simple server socket.
	 * @throws IOException Iff something goes wrong during the socket establishment.
	 * @throws ClassNotFoundException 
	 * @throws SAXException 
	 * @throws ParserConfigurationException 
	 * @throws IllegalAccessException 
	 * @throws InstantiationException 
	 */
	private ServerSocket getPlainServerSocket() throws IOException,
			ParserConfigurationException, SAXException, ClassNotFoundException,
			InstantiationException, IllegalAccessException {
		return new ServerSocket(patternRegistry.getTransport().getPlainServerPort());
	}
	
	/**
	 * Create a secure server socket. Opens a key store specified in the
	 * node-java.xml configuration file, using the credentials and alias
	 * specified therein.
	 * 
	 * @return a server socket.
	 * @throws KeyStoreException Iff something goes wrong.
	 * @throws NoSuchAlgorithmException Iff something goes wrong.
	 * @throws CertificateException Iff something goes wrong.
	 * @throws FileNotFoundException Iff something goes wrong.
	 * @throws IOException Iff something goes wrong.
	 * @throws UnrecoverableKeyException Iff something goes wrong.
	 * @throws KeyManagementException Iff something goes wrong.
	 * @throws ParserConfigurationException Iff something goes wrong.
	 * @throws SAXException Iff something goes wrong.
	 * @throws ClassNotFoundException Iff something goes wrong.
	 * @throws IllegalAccessException 
	 * @throws InstantiationException 
	 */
	private ServerSocket getSecureServerSocket() throws KeyStoreException,
			NoSuchAlgorithmException, CertificateException,
			FileNotFoundException, IOException, UnrecoverableKeyException,
			KeyManagementException, ParserConfigurationException, SAXException,
			ClassNotFoundException, InstantiationException,
			IllegalAccessException {
		Transport trans = patternRegistry.getTransport();
		KeyStore keyStore = KeyStore.getInstance("JKS");
		keyStore.load(getClass().getResourceAsStream(trans.getKeyStorePath()),
				trans.getKeyStorePassword().toCharArray());
		KeyManagerFactory keyMgrFac = KeyManagerFactory.getInstance("SunX509");
		keyMgrFac.init(keyStore, trans.getAliasPassword().toCharArray());
		SSLContext sslContext = SSLContext.getInstance("SSLv3");
		sslContext.init(keyMgrFac.getKeyManagers(), null, null);
		ServerSocketFactory serverSocketFac =
				sslContext.getServerSocketFactory();
		ServerSocket serverSocket =
				serverSocketFac.createServerSocket(trans.getSecureServerPort());
		return serverSocket;
	}
	
	/**
	 * The main loop for the NodeJava system.
	 */
	public void run() {
		try {
			boolean isSecure = patternRegistry.getTransport().isSecure();
			ServerSocket listenSocket = isSecure ?
					getSecureServerSocket() :
					getPlainServerSocket();
			while (true) {
				Socket clientConnection = listenSocket.accept();
				long startTime = System.currentTimeMillis();
				if (logger.isTraceEnabled()) {
					logger.trace("Server socket accepted connection. (" + startTime + ").");
				}
				ProcessClient processClient = processClientPool.borrowObject();
				processClient.init(this, isSecure, startTime,
						listenSocket.getInetAddress().getHostName(),
						listenSocket.getLocalPort(),
						clientConnection);
			}
		} catch (Exception e) {
			logger.error("Exception while waiting for incoming socket connections.", e);
		}
	}
	
	/**
	 * This method will start the node on a simple thread.
	 * 
	 * @param args ignored for now...
	 */
	public static void main(String[] args) {
		try {
			NodeJava node = new NodeJava(null);
			node.run();
		} catch (Exception e) {
			logger.error("Unable to run NodeJava.", e);
		}
	}

}
