package com.koutra.nodejava;

import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;

/**
 * Represents a part in a multipart HTTP form post.
 * 
 * @author pftakas
 */
public interface Part {

	/**
	 * Gets the content of this part as an InputStream.
	 * 
	 * @return The content of this part as an InputStream.
	 * @throws IOException If an error occurs in retrieving the contet as an
	 * InputStream
	 */
	InputStream getInputStream() throws IOException;

	/**
	 * Gets the content type of this part.
	 * 
	 * @return The content type of this part.
	 */
	String getContentType();

	/**
	 * Gets the name of this part.
	 * 
	 * @return The name of this part as a String.
	 */
	String getName();

	/**
	 * Returns the size of this fille.
	 * 
	 * @return a long specifying the size of this part, in bytes.
	 */
	long getSize();

	/**
	 * Returns the value of the specified mime header as a String. If the Part
	 * did not include a header of the specified name, this method returns null.
	 * If there are multiple headers with the same name, this method returns
	 * the first header in the part. The header name is case insensitive. You
	 * can use this method with any request header.
	 * 
	 * @param name a String specifying the header name.
	 * @return a String containing the value of the requested header, or null
	 * if the part does not have a header of that name.
	 */
	String getHeader(String name);

	/**
	 * Gets the values of the Part header with the given name.
	 * 
	 * Any changes to the returned Collection must not affect this Part.
	 * 
	 * Part header names are case insensitive.
	 * 
	 * @param name the header name whose values to return
	 * @return a (possibly empty) Collection of the values of the header with
	 * the given name.
	 */
	Collection<String> getHeaders(String name);

	/**
	 * Gets the header names of this Part.
	 * 
	 * Some servlet containers do not allow servlets to access headers using
	 * this method, in which case this method returns null.
	 * 
	 * Any changes to the returned Collection must not affect this Part.
	 * 
	 * @returna (possibly empty) Collection of the header names of this Part.
	 */
	Collection<String> getHeaderNames();
}
