package com.koutra.nodejava;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;

import javax.xml.parsers.ParserConfigurationException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.SAXException;

import com.koutra.nodejava.impl.HttpRequest;
import com.koutra.nodejava.impl.HttpResponse;

/**
 * A helper class used to handle the mechanics of handling the client socket
 * and all the relevant host name, port number, remote information etc. The
 * NodeJava class is using this class to handle a particular client conversation
 * for as long as the connection is kept alive.
 * 
 * In turn this class, after abstracting all the socket relevant information,
 * delegates to the ProcessClientTransportAgnostic class to perform the actual
 * processing. Although this object might potentially handle a number of HTTP
 * requests (assuming the 'keep-alive' flag is used by the client), one instance
 * of the ProcessClientTransportAgnostic class will be used per request.
 * 
 * @author pftakas
 */
public class ProcessClient extends Thread implements TimeRecorder {
	private static final Logger logger = LoggerFactory.getLogger(ProcessClient.class);
	
	protected boolean initialized;
	protected NodeJava nodeJava;
	protected boolean secure;
	protected long startTime;
	protected Socket clientSocket;
	protected String serverName;
	protected int serverPort;
	
	public ProcessClient() {
		this.initialized = false;
	}
	
	/**
	 * Initializing method.
	 * @param nodeJava the NodeJava instance that this process client belongs to.
	 * @param secure true iff the processing client is over a secure socket.
	 * @param startTime the time the socket was opened.
	 * @param serverName the name of the server host.
	 * @param serverPort the server port number.
	 * @param clientSocket the socket opened for the communication.
	 * @throws ParserConfigurationException iff something goes wrong.
	 * @throws SAXException iff something goes wrong.
	 * @throws IOException iff something goes wrong.
	 * @throws ClassNotFoundException iff something goes wrong.
	 * @throws NodeJavaException 
	 */
	public void init(NodeJava nodeJava, boolean secure, long startTime,
			String serverName, int serverPort, Socket clientSocket)
					throws ParserConfigurationException, SAXException,
					IOException, ClassNotFoundException, NodeJavaException {
		if (initialized) {
			throw new NodeJavaException("Unexpected process client state: " +
					"Trying to initialize an already initialized instance.");
		}
		
		this.initialized = true;
		this.nodeJava = nodeJava;
		this.secure = secure;
		this.startTime = startTime;
		this.serverName = serverName;
		this.serverPort = serverPort;
		this.clientSocket = clientSocket;

		nodeJava.getProcessClientExecutorService().execute(this);
	}
	
	/**
	 * Method to set the state to uninitialized and release the instance back
	 * to the pool.
	 * @throws Exception 
	 */
	protected void releaseOurselves() throws Exception {
		if (!initialized) {
			throw new NodeJavaException("Unexpected process client state: " +
					"Trying to release a non-initialized instance.");
		}
		
		this.initialized = false;
		this.secure = false;
		this.startTime = 0L;
		this.serverName = null;
		this.serverPort = 0;
		this.clientSocket = null;
		
		nodeJava.releaseProcessClient(this);
	}

	@Override
	/**
	 * Helper method that sets the start time. Used in the case where the keep
	 * alive flag is set to true, and we are waiting to receive the first
	 * byte from the other side, before starting counting the time. Called by
	 * the underlying instance of ProcessClientTransportAgnostic.
	 * @param startTime the time to start counting for the next thread.
	 */
	public void setStartTime(long startTime) {
		this.startTime = startTime;
	}
	
	@Override
	/**
	 * The main method of the thread that is executing this client. The main
	 * loop repeats while the requests specify the keep alive header flag.
	 * For each iteration, it creates an instance of ProcessClientTransportAgnostic,
	 * reads the request headers with its help, construct the response, and
	 * delegates the processing of the request to the instance of
	 * ProcessClientTransportAgnostic. Also the logic of managing the time
	 * counting  for request processing is implemented in this method.
	 */
	public void run() {
		try {
			String remoteAddr = clientSocket.getInetAddress().getHostAddress();
			String remoteHost = clientSocket.getInetAddress().getHostName();
			InputStream socketInputStream = clientSocket.getInputStream();
			OutputStream socketOutputStream = clientSocket.getOutputStream();
			
			ProcessClientTransportAgnostic processClientProper =
					nodeJava.acquireProcessClientTransport();
			try {
				processClientProper.init(
								nodeJava, this, secure, remoteAddr, remoteHost,
								serverName, serverPort, socketInputStream,
								socketOutputStream);
				boolean firstTime = true;
				while (true) {
					if (firstTime) {
						// The start time is passed through the constructor to
						// take into account the ProcessClient and thread start
						// times.
						firstTime = false;
					} else {
						// We are keeping the socket alive, so the processing
						// time will start when we receive the first byte from
						// the client.
						startTime = 0L;
					}
					if (logger.isTraceEnabled()) {
						logger.trace("Waiting for incoming HTTP request.");
					}
	
					HttpRequest request = processClientProper.processHeaders();
					if (logger.isTraceEnabled()) {
						logger.trace("HTTP request headers received.");
					}
					HttpResponse response = new HttpResponse(socketOutputStream);
					processClientProper.processRequest(request, response);
					response.terminate();
					
					long endTime = System.currentTimeMillis();
					if (logger.isTraceEnabled()) {
						logger.trace("HTTP request processed. (" + endTime + ").");
					}
					
					if (logger.isTraceEnabled()) {
						logger.trace("Request processing took: " +
								(endTime - startTime)  + "ms.");
					}
					
					if (!request.isKeepAlive()) {
						if (logger.isTraceEnabled()) {
							logger.trace("HTTP request could not be kept alive. Exiting thread.");
						}
						clientSocket.close();
						break;
					}
				}
			} finally {
				nodeJava.releaseProcessClientTransport(processClientProper);
			}
		} catch (Exception e) {
			logger.error("Error while processing request. " +
					"Processing loop exiting.", e);
		} finally {
			try {
				releaseOurselves();
			} catch (Exception e) {
				logger.error("Error while trying to release process client back to the pool.", e);
			}
		}
	}
}
