package com.koutra.nodejava;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.xml.parsers.ParserConfigurationException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.SAXException;

import com.koutra.nodejava.impl.HttpRequest;
import com.koutra.nodejava.impl.HttpResponse;
import com.koutra.nodejava.registry.PatternRegistry;
import com.koutra.nodejava.utils.HeaderReadingInputStream;
import com.koutra.nodejava.utils.HttpConstants;

/**
 * Helper class that contains the logic of constructing the HttpRequest
 * object out of the contents of the input stream that is passed to it as
 * a parameter during construction.
 * 
 * It delegates actual processing of the request to the PatternRegistry
 * singleton.
 * 
 * @author pftakas
 */
public class ProcessClientTransportAgnostic {
	private static final Logger logger = LoggerFactory.getLogger(
			ProcessClientTransportAgnostic.class);
	
	protected boolean initialized;
	protected boolean secure;
	protected String remoteAddr;
	protected String remoteHost;
	protected String serverName;
	protected int serverPort;
	protected NodeJava nodeJava;
	protected TimeRecorder timeRecorder;
	protected InputStream inputStream;
	protected OutputStream outputStream;
	protected PatternRegistry registry;
	
	/**
	 * Initializing constructor.
	 */
	public ProcessClientTransportAgnostic() {
		this.initialized = false;
	}
	
	/**
	 * Initializing method.
	 * @param timeRecorder handle to the callback to set the start time of the
	 * request processing.
	 * @param secure true iff the processing client is over a secure socket.
	 * @param remoteAddr the remote network address.
	 * @param remoteHost the remote host name.
	 * @param serverName the server host name.
	 * @param serverPort the server port number.
	 * @param inputStream the socket input stream.
	 * @param outputStream the socket output stream.
	 * @throws ParserConfigurationException iff something goes wrong.
	 * @throws SAXException iff something goes wrong.
	 * @throws IOException iff something goes wrong.
	 * @throws ClassNotFoundException iff something goes wrong.
	 * @throws IllegalAccessException iff something goes wrong.
	 * @throws InstantiationException iff something goes wrong.
	 */
	public void init(NodeJava nodeJava, TimeRecorder timeRecorder,
			boolean secure, String remoteAddr, String remoteHost,
			String serverName, int serverPort, InputStream inputStream,
			OutputStream outputStream) throws ParserConfigurationException,
			SAXException, IOException, ClassNotFoundException,
			InstantiationException, IllegalAccessException {
		this.nodeJava = nodeJava;
		this.timeRecorder = timeRecorder;
		this.secure = secure;
		this.remoteAddr = remoteAddr;
		this.remoteHost = remoteHost;
		this.serverName = serverName;
		this.serverPort = serverPort;
		this.inputStream = new BufferedInputStream(inputStream); // add a buffer for performance
		this.outputStream = outputStream;
		this.registry = nodeJava.getPatternRegistry();
	}
	
	/**
	 * Helper method used by the header reading input stream to set the starting
	 * time of a request processing when the first byte is received.
	 * @param startTime the time in milliseconds since the start of the epoch,
	 * when the first request byte was received by the server.
	 */
	public void setStartTime(long startTime) {
		timeRecorder.setStartTime(startTime);
	}
	
	/**
	 * Process the headers of an incoming request constructing the HttpRequest
	 * along the way.
	 * @return the constructed HttpRequest.
	 * @throws IOException iff something goes wrong in the socket communication.
	 * @throws NodeJavaException iff something wrong in the HTTP communication
	 * (e.g. malformed HTTP header received).
	 */
	protected HttpRequest processHeaders() throws IOException, NodeJavaException {
		HeaderReadingInputStream headerReader =
				new HeaderReadingInputStream(this, inputStream);
		HttpRequest request = new HttpRequest(secure, serverName, serverPort,
				remoteAddr, remoteHost, inputStream);
		boolean firstLineProcessed = false;
		while (true) {
			String headerLine = headerReader.readHeaderLine(!firstLineProcessed);
			if (logger.isTraceEnabled()) {
				logger.trace("RECEIVED: " + headerLine);
			}
			if (headerLine.length() == 0) {
				// We have reached the end of the HTTP headers.
				// Check if we have any optional part (POST request data, multi
				// part data, etc.)
				request.processPostHeaderData();
				break;
			}
			
			if (!firstLineProcessed) {
				processFirstLine(headerLine, request);
				firstLineProcessed = true;
			} else {
				int colonIndex = headerLine.indexOf(':');
				String headerName = headerLine.substring(0, colonIndex);
				String headerValue = headerLine.substring(colonIndex + 1).trim();
				request.addHeader(headerName, headerValue);
			}
		}
		return request;
	}
	
	/**
	 * Helper method to process the first line of an incoming HTTP request.
	 * @param line the line read from the socket.
	 * @param request the HttpRequest to initialize.
	 * @throws NodeJavaException iff something goes wrong (e.g. malformed
	 * HTTP request first line).
	 */
	protected void processFirstLine(String line, HttpRequest request)
			throws NodeJavaException {
		// First deal with the protocol.
		int lastSpaceCharIndex = line.lastIndexOf(' ');
		if (lastSpaceCharIndex == -1) {
			throw new NodeJavaException(
					"Malformed HTTP request line: '" + line + "'.");
		}
		
		String protocol = line.substring(lastSpaceCharIndex + 1);
		String methodAndURI = line.substring(0, lastSpaceCharIndex);
		if (!HttpConstants.PROTOCOL_HTTP_1_0.equals(protocol) &&
				!HttpConstants.PROTOCOL_HTTP_1_1.equals(protocol)) {
			throw new NodeJavaException(
					"Unknown HTTP protocol string '" + protocol + "'.");
		}
		
		request.setProtocol(protocol);
		
		int firstSpaceCharIndex = methodAndURI.indexOf(' ');
		if (firstSpaceCharIndex == -1) {
			throw new NodeJavaException(
					"Malformed HTTP request line: '" + line + "'.");
		}
		
		String method = methodAndURI.substring(0, firstSpaceCharIndex);
		String uri = methodAndURI.substring(firstSpaceCharIndex + 1);
		if (!HttpConstants.METHOD_OPTIONS.equals(method) &&
				!HttpConstants.METHOD_GET.equals(method) &&
				!HttpConstants.METHOD_HEAD.equals(method) &&
				!HttpConstants.METHOD_POST.equals(method) &&
				!HttpConstants.METHOD_PUT.equals(method) &&
				!HttpConstants.METHOD_DELETE.equals(method) &&
				!HttpConstants.METHOD_TRACE.equals(method) &&
				!HttpConstants.METHOD_CONNECT.equals(method)) {
			throw new NodeJavaException(
					"Unknown HTTP method '" + method + "'.");
		}
		
		request.setMethod(method);
		
		int questionMarkIndex = uri.indexOf('?');
		if (questionMarkIndex == -1) {
			request.setPath(uri);
		} else {
			request.setPath(uri.substring(0, questionMarkIndex));
			request.setQueryString(uri.substring(questionMarkIndex + 1));
		}
	}
	
	/**
	 * Process the request. Delegates the processing to the PatternRegistry
	 * instance.
	 * @param request the HttpRequest.
	 * @param response the HttpResponse.
	 * @throws NodeJavaException iff something goes wrong during the processing.
	 */
	protected void processRequest(HttpRequest request, HttpResponse response)
			throws NodeJavaException {
		registry.process(request, response);
	}
	
}
