package com.koutra.nodejava;

/**
 * A specialization of the HttpProcessor that is meant to be used by the end
 * users in order to process HTTP requests. Inheritors of this class
 * should make every attempt to keep themselves stateless, because instances
 * of their classes will be kept by the NodeJava implementation in object
 * pools maintained by the system and re-used accross requests.
 * 
 * The carefull reader might note that even though this class is named
 * 'Stateless' it has a state: the node java instance under which this
 * processor is running. The irony is not lost on the writer but the stateless
 * part is intended as a hint to inheritors, not as an iron clad requirement.
 * 
 * @author pftakas
 */
public abstract class StatelessHttpProcessor implements HttpProcessor {
	protected NodeJava nodeJava;
	
	/**
	 * Callback to notify the processor of the NodeJava instance it is running
	 * on. Implementors can override this to register facilities with the
	 * NodeJava node, so that they do not keep state themselves. Implementors
	 * must call super.init() when overriding.
	 * @param nodeJava the NodeJava instance on which the processor is running.
	 */
	public void init(NodeJava nodeJava) throws NodeJavaException {
		this.nodeJava = nodeJava;
	}
}
