package com.koutra.nodejava.impl;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.nio.charset.Charset;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.koutra.nodejava.Cookie;
import com.koutra.nodejava.HttpSession;
import com.koutra.nodejava.NodeJavaException;
import com.koutra.nodejava.Part;
import com.koutra.nodejava.registry.PatternRegistry;
import com.koutra.nodejava.utils.DateHelper;
import com.koutra.nodejava.utils.HttpConstants;
import com.koutra.nodejava.utils.MultipartStreamReader;

public class HttpRequest implements com.koutra.nodejava.HttpRequest {
	private static final Logger logger = LoggerFactory.getLogger(HttpRequest.class);

	@Override
	public Object getAttribute(String name) {
		return attributeMap.get(name);
	}

	@Override
	public Set<String> getAttributeNames() {
		return attributeMap.keySet();
	}

	@Override
	public void setAttribute(String name, Object value) {
		attributeMap.put(name, value);
	}

	@Override
	public void removeAttribute(String name) {
		attributeMap.remove(name);
	}

	@Override
	public String getCharacterEncoding() {
		return characterEncoding;
	}

	@Override
	public void setCharacterEncoding(String enc)
			throws UnsupportedEncodingException {
		this.characterEncoding = enc;
	}

	@Override
	public int getContentLength() {
		return contentLength;
	}

	@Override
	public String getContentType() {
		return contentType;
	}

	@Override
	public InputStream getInputStream() throws IOException {
		if (readerRequested) {
			throw new IllegalStateException(
					"getInputStream() in being called, but getReader() " +
					"has already been called."); 
		}
		
		inputStreamRequested = true;
		return requestInputStream;
	}

	@Override
	public BufferedReader getReader() throws IOException {
		if (inputStreamRequested) {
			throw new IllegalStateException(
					"getReader() in being called, but getInputStream() " +
					"has already been called."); 
		}
		
		readerRequested = true;
		Charset charset = Charset.forName(characterEncoding);
		return new BufferedReader(new InputStreamReader(
				requestInputStream, charset));
	}

	@Override
	public String getParameter(String name) {
		String[] paramValues = parameterMap.get(name);
		if (paramValues == null || paramValues.length == 0) {
			return null;
		}
		return paramValues[0];
	}

	@Override
	public Set<String> getParameterNames() {
		return parameterMap.keySet();
	}

	@Override
	public String[] getParameterValues(String name) {
		String[] paramValues = parameterMap.get(name);
		if (paramValues == null) {
			return null;
		}
		return Arrays.copyOf(paramValues, paramValues.length);
	}

	@Override
	public Map<String, String[]> getParameterMap() {
		return Collections.unmodifiableMap(parameterMap);
	}

	@Override
	public String getProtocol() {
		return protocol;
	}

	@Override
	public String getScheme() {
		return scheme;
	}

	@Override
	public String getServerName() {
		return serverName;
	}

	@Override
	public int getServerPort() {
		return serverPort;
	}

	@Override
	public String getRemoteAddr() {
		return remoteAddr;
	}

	@Override
	public String getRemoteHost() {
		return remoteHost;
	}

	@Override
	public Locale getLocale() {
		if (localeList == null || localeList.size() == 0) {
			return Locale.getDefault();
		}
		return localeList.get(0);
	}

	@Override
	public List<Locale> getLocales() {
		return Collections.unmodifiableList(localeList);
	}

	@Override
	public boolean isSecure() {
		return secure;
	}

	@Override
	public Cookie[] getCookies() {
		return cookieList.toArray(new Cookie[cookieList.size()]);
	}

	@Override
	public long getDateHeader(String name) {
		String value = getHeader(name);
		if (value == null) {
			return -1;
		}
		try {
			Date date = DateHelper.getRfc1123Format().parse(value);
			return date.getTime();
		} catch (ParseException e) {
			throw new IllegalArgumentException("Unable to covert '" + value +
					"' to an RFC1123 date.", e);
		}
	}

	@Override
	public String getHeader(String name) {
		String[] headerValues = headerMap.get(name);
		if (headerValues == null || headerValues.length == 0) {
			return null;
		}
		return headerValues[0];
	}

	@Override
	public List<String> getHeaders(String name) {
		ArrayList<String> returnValue = new ArrayList<String>();
		String[] headerValues = headerMap.get(name);
		for (String headerValue : headerValues) {
			returnValue.add(headerValue);
		}
		return returnValue;
	}

	@Override
	public Set<String> getHeaderNames() {
		return headerMap.keySet();
	}

	@Override
	public int getIntHeader(String name) {
		String value = getHeader(name);
		return Integer.parseInt(value);
	}

	@Override
	public String getMethod() {
		return method;
	}

	@Override
	public String getPath() {
		return path;
	}

	@Override
	public String getQueryString() {
		return queryString;
	}

	@Override
	public String getRequestURI() {
		return path;
	}

	@Override
	public StringBuffer getRequestURL() {
		StringBuffer returnValue = new StringBuffer();
		returnValue.append(scheme);
		returnValue.append("://");
		returnValue.append(serverName);
		if ((HttpConstants.HTTP_SCHEME.equals(scheme) && serverPort != 80) ||
			(HttpConstants.HTTPS_SCHEME.equals(scheme) && serverPort != 443)) {
			returnValue.append(":");
			returnValue.append(serverPort);
		}
		returnValue.append(path);
		return returnValue;
	}

	@Override
	public HttpSession getSession(boolean create) {
		if (session == null) {
			// First try to get the session cookie and see if we can retrieve
			// a valid session from the SessionContainer.
			for (Cookie cookie : getCookies()) {
				if (HttpConstants.COOKIE_SESSION_NAME.equals(cookie.getName())) {
					String sessionId = cookie.getValue();
					session = SessionContainer.get().getSession(sessionId);
					if (session == null || !session.isValid()) {
						SessionContainer.get().removeSession(sessionId);
						session = null;
					}
				}
			}
		}
		
		if (session == null && create) {
			session = new com.koutra.nodejava.impl.HttpSession();
			try {
				session.setMaxInactiveInterval(PatternRegistry.get().getSession().getTimeout());
			} catch (Exception e) {
				logger.error("Unable to get the session timeout value.", e);
				throw new IllegalStateException(
						"Unable to get the session timeout value.", e);
			}
		}
		
		return session;
	}

	@Override
	public HttpSession getSession() {
		return getSession(true);
	}

	@Override
	public Collection<Part> getParts() throws IOException, NodeJavaException {
		if (!isMultipartContent()) {
			throw new NodeJavaException("Trying to retrieve multipart " +
					"encoding parts, on a request that is multipart encoded.");
		}
		
		if (!multipartParsed) {
			parseMultipartEncodedContents();
		}
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Part getPart(String name) throws IOException, NodeJavaException {
		if (!isMultipartContent()) {
			throw new NodeJavaException("Trying to retrieve a multipart " +
					"encoding part, on a request that is multipart encoded.");
		}
		
		// TODO Auto-generated method stub
		return null;
	}
	
	////////////////////////////////////////////////////////////////////////////
	// End of interface implementation. Now for the specifics.
	////////////////////////////////////////////////////////////////////////////
	
	protected String protocol;
	protected String scheme;
	protected int contentLength;
	protected String contentType;
	protected String characterEncoding;
	protected String serverName;
	protected int serverPort;
	protected String remoteAddr;
	protected String remoteHost;
	protected String method;
	protected String path;
	protected String queryString;
	protected Map<String, String[]> parameterMap;
	protected Map<String, Object> attributeMap;
	protected List<Locale> localeList;
	protected boolean secure;
	protected List<Cookie> cookieList;
	protected Map<String, String[]> headerMap;
	protected HttpSession session;
	protected InputStream requestInputStream;
	protected boolean inputStreamRequested;
	protected boolean readerRequested;
	protected boolean keepAlive;
	protected boolean multipartParsed;
	protected Map<String, Part> partMap;
	
	public HttpRequest(boolean secure, String serverName, int serverPort,
			String remoteAddr, String remoteHost, InputStream inputStream) {
		this.protocol = null;
		this.scheme = null;
		this.contentLength = -1;
		this.contentType = null;
		this.characterEncoding = null;
		this.serverName = serverName;
		this.serverPort = serverPort;
		this.remoteAddr = remoteAddr;
		this.remoteHost = remoteHost;
		this.method = null;
		this.path = null;
		this.queryString = null;
		this.parameterMap = new HashMap<String, String[]>();
		this.attributeMap = new HashMap<String, Object>();
		this.localeList = new ArrayList<Locale>();
		this.secure = secure;
		this.cookieList = new ArrayList<Cookie>();
		this.headerMap = new HashMap<String, String[]>();
		this.session = null;
		this.requestInputStream = inputStream;
		this.inputStreamRequested = false;
		this.readerRequested = false;
		this.keepAlive = false;
		this.multipartParsed = false;
		this.partMap = null;
	}

	public void setProtocol(String protocol) {
		this.protocol = protocol;
	}

	public void setScheme(String scheme) {
		this.scheme = scheme;
	}

	public void setContentLength(int contentLength) {
		this.contentLength = contentLength;
	}

	public void setContentType(String contentType) {
		this.contentType = contentType;
	}

	public void setMethod(String method) {
		this.method = method;
	}

	public void setPath(String path) {
		this.path = path;
	}
	
	protected void addCookieValues(String headerValue) {
		String[] cookieParts = headerValue.split(";");
		for (String cookiePart : cookieParts) {
			int equalsIndex = cookiePart.indexOf('=');
			if (equalsIndex == -1) {
				logger.warn(
						"Malformed name value pair in cookie string: '" +
								cookiePart + "'.");
				cookieList.add(new Cookie(cookiePart, ""));
			} else {
				String name = cookiePart.substring(0, equalsIndex);
				String value = cookiePart.substring(equalsIndex + 1);
				cookieList.add(new Cookie(name, value));
			}
		}
	}
	
	public void addHeader(String name, String value) {
		if (HttpConstants.HEADER_NAME_ACCEPT_LANGUAGE.equals(name)) {
			String[] values = value.split("\\,");
			for (String langString : values) {
				String lang = langString.toLowerCase();
				String country = null;
				if (langString.length() != 2) {
					lang = langString.substring(0, 2).toLowerCase();
					country = langString.substring(3, 5).toUpperCase();
					localeList.add(new Locale(lang, country));
				} else {
					localeList.add(new Locale(lang));
				}
			}
		} else if (HttpConstants.HEADER_NAME_COOKIE.equals(name)) {
			addCookieValues(value);
		} else if (HttpConstants.HEADER_NAME_CONNECTION.equals(name)) {
			if (HttpConstants.HEADER_VALUE_KEEP_ALIVE.equals(value)) {
				keepAlive = true;
			}
		}
		
		String[] values = headerMap.get(name);
		if (values == null) {
			if (HttpConstants.HEADER_NAME_ACCEPT.equals(name)) {
				setContentType(value);
			} else if (HttpConstants.HEADER_NAME_CONTENT_LENGTH.equals(name)) {
				try {
					setContentLength(Integer.parseInt(value));
				} catch (NumberFormatException e) {
					logger.warn("Malformed Content-Length header value: '" +
							value + "'.");
				}
			}
			headerMap.put(name, new String[] { value });
			return;
		}
		
		String[] modifiedArray = Arrays.copyOf(values, values.length + 1);
		modifiedArray[values.length] = value;
		
		headerMap.put(name, modifiedArray);
	}
	
	protected void addParameter(String name, String value) {
		String[] values = parameterMap.get(name);
		if (values == null) {
			parameterMap.put(name, new String[] { value });
			return;
		}
		
		String[] modifiedArray = Arrays.copyOf(values, values.length + 1);
		modifiedArray[values.length] = value;
		
		parameterMap.put(name, modifiedArray);
	}

	public void setQueryString(String queryString) {
		this.queryString = queryString;
		if (queryString != null && !queryString.isEmpty()) {
			String[] nameValuePairs = queryString.split("&");
			for (String nameValuePair : nameValuePairs) {
				int equalsIndex = nameValuePair.indexOf('=');
				if (equalsIndex == -1) {
					logger.warn(
							"Malformed name value pair in query string: '" +
									nameValuePair + "'.");
					addParameter(nameValuePair, "");
				} else {
					String name = nameValuePair.substring(0, equalsIndex);
					String value = nameValuePair.substring(equalsIndex + 1);
					try {
						addParameter(name, URLDecoder.decode(value, "UTF-8"));
					} catch (UnsupportedEncodingException e) {
						logger.warn("Unable to URL decode: '" + value + "'.");
					}
				}
			}
		}
	}

	public void setSession(HttpSession session) {
		this.session = session;
	}
	
	public boolean isKeepAlive() {
		return keepAlive;
	}
	
	public void processPostHeaderData() throws IOException {
		if (headerMap.containsKey(HttpConstants.HEADER_NAME_CONTENT_TYPE) &&
				headerMap.get("Content-Type")[0].equals(
						HttpConstants.HEADER_VALUE_FORM_ENCODED)) {
			// This is a POST from a form. We must parse the form data.
			int contentLength = Integer.parseInt(
					headerMap.get(HttpConstants.HEADER_NAME_CONTENT_LENGTH)[0]);
			byte[] postParams = new byte[contentLength];
			readAllPostBytes(postParams);
			String parameters = new String(postParams, "UTF-8");
			String[] params = parameters.split("&");
			for (String nameValuePair : params) {
				int equalsIndex = nameValuePair.indexOf('=');
				if (equalsIndex == -1) {
					logger.warn(
							"Malformed name value pair in query string: '" +
									nameValuePair + "'.");
					addParameter(nameValuePair, "");
				} else {
					String name = nameValuePair.substring(0, equalsIndex);
					String value = nameValuePair.substring(equalsIndex + 1);
					try {
						addParameter(name, URLDecoder.decode(value, "UTF-8"));
					} catch (UnsupportedEncodingException e) {
						logger.warn("Unable to URL decode: '" + value + "'.");
					}
				}
			}
		}
	}
	
	private void readAllPostBytes(byte[] postParams) throws IOException {
		int offset = 0;
		int len = postParams.length;
		while (true) {
			int readBytes = requestInputStream.read(postParams, offset, len);
			offset += readBytes;
			len -= readBytes;
			if (len == 0) {
				break;
			}
		}
	}
	
	private boolean isMultipartContent() {
		if (contentType == null) {
			return false;
		}
		
		if (contentType.toLowerCase().startsWith(HttpConstants.HEADER_VALUE_MULTIPART_START)) {
			return true;
		}
		
		return false;
	}
	
	private void parseMultipartEncodedContents() throws NodeJavaException, IOException {
		String boundary = null;
		String[] nameValueStrings = contentType.split(";");
		for (int i = 0; i < nameValueStrings.length; i++) {
			String nameValue = nameValueStrings[i];
			int equalsIndex = nameValue.indexOf('=');
			if (equalsIndex < 0) {
				continue;
			}
			String name = nameValue.substring(0, equalsIndex);
			String value = nameValue.substring(equalsIndex + 1);
			if (name.equals("boundary")) {
				boundary = value;
				break;
			}
		}
		
		if (boundary == null) {
			throw new NodeJavaException("No boundary specified for " +
					"multipart encoded request.");
		}
		
		MultipartStreamReader reader =
				new MultipartStreamReader(getInputStream(), boundary);
		partMap = reader.parse();
		multipartParsed = true;
	}
}
