package com.koutra.nodejava.impl;

import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.koutra.nodejava.Cookie;
import com.koutra.nodejava.utils.ByteCountingOutputStream;
import com.koutra.nodejava.utils.ByteCountingPrintWriter;
import com.koutra.nodejava.utils.ChunkedEncodingOutputStream;
import com.koutra.nodejava.utils.ChunkedEncodingPrintWriter;
import com.koutra.nodejava.utils.DateHelper;
import com.koutra.nodejava.utils.HttpConstants;
import com.koutra.nodejava.utils.ResponseAwareOutputStream;
import com.koutra.nodejava.utils.ResponseAwarePrintWriter;

public class HttpResponse implements com.koutra.nodejava.HttpResponse {
	
	@Override
	public String getCharacterEncoding() {
		return characterEncoding;
	}

	@Override
	public String getContentType() {
		return contentType;
	}
	
	@Override
	public void setChunkedEncoding(boolean chunkedEncoding) {
		if (commited) {
			throw new IllegalStateException(
					"setChunkedEncoding() is being called on an already commited response.");
		}
		
		if (writerRequested || outputStreamRequested) {
			throw new IllegalStateException(
					"setChunkedEncoding() is being called, but a " +
					"reader/writer has already been requested.");
		}

		isUsingChunkedEncoding = chunkedEncoding;
	}

	@Override
	public OutputStream getOutputStream() throws IOException {
		if (writerRequested) {
			throw new IllegalStateException(
					"getOutputStream() is being called, but getWriter() " +
					"has already been called."); 
		}
		
		if (!outputStreamRequested) {
			outputStreamRequested = true;
			if (isUsingChunkedEncoding) {
				responseOutputStream = new ChunkedEncodingOutputStream(
						new BufferedOutputStream(new ResponseAwareOutputStream(
								this, originalOutputStream),
						bufferSize), bufferSize);
			} else {
				responseOutputStream = new ByteCountingOutputStream(
						new BufferedOutputStream(new ResponseAwareOutputStream(
								this, originalOutputStream),
						bufferSize));
			}
		}
		return responseOutputStream;
	}

	@Override
	public PrintWriter getWriter() throws IOException {
		if (outputStreamRequested) {
			throw new IllegalStateException(
					"getWriter() is being called, but getOutputStream() " +
					"has already been called."); 
		}
		
		if (!writerRequested) {
			writerRequested = true;
			// TODO correctly support character encoding.
			// Charset charset = Charset.forName(characterEncoding);
			Charset charset = Charset.forName("UTF-8");
			if (isUsingChunkedEncoding) {
				responsePrintWriter = new ChunkedEncodingPrintWriter(new BufferedWriter(
						new ResponseAwarePrintWriter(
								this, new OutputStreamWriter(
										originalOutputStream, charset))
						, bufferSize), "UTF-8", bufferSize);
			} else {
				responsePrintWriter = new ByteCountingPrintWriter(new BufferedWriter(
						new ResponseAwarePrintWriter(
								this, new OutputStreamWriter(
										originalOutputStream, charset))
						, bufferSize), "UTF-8");
			}
		}
		return responsePrintWriter;
	}

	@Override
	public void setCharacterEncoding(String charset) {
		characterEncoding = charset;
	}

	@Override
	public void setContentLength(int len) {
		contentLength = len;
	}

	@Override
	public void setContentType(String type) {
		contentType = type;
	}

	@Override
	public void setBufferSize(int size) {
		if (commited ||
				responseOutputStream != null ||
				responsePrintWriter != null) {
			throw new IllegalStateException(
					"setBufferSize() invoked on a commited response.");
		}

		bufferSize = size;
	}

	@Override
	public int getBufferSize() {
		return bufferSize;
	}

	@Override
	public void flushBuffer() throws IOException {
		if (!commited) {
			commit();
			commited = true;
		}
		
		if (responsePrintWriter != null) {
			responsePrintWriter.flush();
		} else {
			if (responseOutputStream == null) {
				getOutputStream();
			}
			responseOutputStream.flush();
		}
	}
	
	@Override
	public void terminate() throws IOException {
		flushBuffer();
		
		if (isUsingChunkedEncoding) {
			if (responseOutputStream != null) {
				((ChunkedEncodingOutputStream) responseOutputStream).terminateStream();
			}
			if (responsePrintWriter != null) {
				((ChunkedEncodingPrintWriter) responsePrintWriter).terminateWriter();
			}
		}
	}

	@Override
	public void resetBuffer() {
		if (commited) {
			throw new IllegalStateException("setBufferSize() invoked on a commited response.");
		}
		
		responsePrintWriter = null;
		responseOutputStream = null;
		outputStreamRequested = false;
		writerRequested = false;
	}

	@Override
	public boolean isCommitted() {
		return commited;
	}

	@Override
	public void reset() {
		if (commited) {
			throw new IllegalStateException("setBufferSize() invoked on a commited response.");
		}
		
		resetBuffer();
		status = -1;
		headerMap.clear();
	}

	@Override
	public void setLocale(Locale loc) {
		this.locale = loc;
	}

	@Override
	public Locale getLocale() {
		return locale;
	}

	@Override
	public void addCookie(Cookie cookie) {
		cookieList.add(cookie);
	}

	@Override
	public boolean containsHeader(String name) {
		return headerMap.containsKey(name);
	}

	@Override
	public String encodeURL(String url) {
		// TODO Auto-generated method stub
		return url;
	}

	@Override
	public String encodeRedirectURL(String url) {
		// TODO Auto-generated method stub
		return url;
	}

	@Override
	public void sendError(int sc, String msg) throws IOException {
		if (commited) {
			throw new IllegalStateException("sendError() invoked on a commited response.");
		}

		// TODO Auto-generated method stub

	}

	@Override
	public void sendError(int sc) throws IOException {
		if (commited) {
			throw new IllegalStateException("sendError() invoked on a commited response.");
		}

		setStatus(sc);
		headerMap.clear();
		commit();
	}

	@Override
	public void sendRedirect(String location) throws IOException {
		if (commited) {
			throw new IllegalStateException("sendRedirect() invoked on a commited response.");
		}
		
		setStatus(SC_TEMPORARY_REDIRECT);
		headerMap.clear();
		headerMap.put(HttpConstants.HEADER_NAME_LOCATION,
				new String[] { encodeRedirectURL(location) } );
		commit();
	}

	@Override
	public void setDateHeader(String name, long date) {
		headerMap.put(name, new String[] {
				DateHelper.getRfc1123Format().format(new Date(date))
				});
	}

	@Override
	public void addDateHeader(String name, long date) {
		String[] values = headerMap.get(name);
		if (values == null) {
			headerMap.put(name, new String[] {
					DateHelper.getRfc1123Format().format(new Date(date))
					});
			return;
		}
		
		String[] modifiedValues = Arrays.copyOf(values, values.length + 1);
		modifiedValues[values.length] =
				DateHelper.getRfc1123Format().format(new Date(date));
		headerMap.put(name, modifiedValues);
	}

	@Override
	public void setHeader(String name, String value) {
		headerMap.put(name, new String[] { value });
	}

	@Override
	public void addHeader(String name, String value) {
		String[] values = headerMap.get(name);
		if (values == null) {
			headerMap.put(name, new String[] { value });
			return;
		}
		
		String[] modifiedValues = Arrays.copyOf(values, values.length + 1);
		modifiedValues[values.length] = value;
		headerMap.put(name, modifiedValues);
	}

	@Override
	public void setIntHeader(String name, int value) {
		headerMap.put(name, new String[] { Integer.toString(value) });
	}

	@Override
	public void addIntHeader(String name, int value) {
		String[] values = headerMap.get(name);
		if (values == null) {
			headerMap.put(name, new String[] { Integer.toString(value) });
			return;
		}
		
		String[] modifiedValues = Arrays.copyOf(values, values.length + 1);
		modifiedValues[values.length] = Integer.toString(value);
		headerMap.put(name, modifiedValues);
	}

	@Override
	public void setStatus(int sc) {
		this.status = sc;
	}

	////////////////////////////////////////////////////////////////////////////
	// End of interface implementation. Now for the specifics.
	////////////////////////////////////////////////////////////////////////////
	private static final Logger logger = LoggerFactory.getLogger(HttpResponse.class);
	
    protected static final Map<Integer, String> statusCodeMap =
    		new HashMap<Integer, String>();
    
    static {
        // HTTP 1.0 Server status codes -- see RFC 1945
        addStatusCodeMap(SC_OK, "OK");
        addStatusCodeMap(SC_CREATED, "Created");
        addStatusCodeMap(SC_ACCEPTED, "Accepted");
        addStatusCodeMap(SC_NO_CONTENT, "No Content");
        addStatusCodeMap(SC_MOVED_PERMANENTLY, "Moved Permanently");
        addStatusCodeMap(SC_MOVED_TEMPORARILY, "Moved Temporarily");
        addStatusCodeMap(SC_NOT_MODIFIED, "Not Modified");
        addStatusCodeMap(SC_BAD_REQUEST, "Bad Request");
        addStatusCodeMap(SC_UNAUTHORIZED, "Unauthorized");
        addStatusCodeMap(SC_FORBIDDEN, "Forbidden");
        addStatusCodeMap(SC_NOT_FOUND, "Not Found");
        addStatusCodeMap(SC_INTERNAL_SERVER_ERROR,
                "Internal Server Error");
        addStatusCodeMap(SC_NOT_IMPLEMENTED, "Not Implemented");
        addStatusCodeMap(SC_BAD_GATEWAY, "Bad Gateway");
        addStatusCodeMap(SC_SERVICE_UNAVAILABLE, "Service Unavailable");

        // HTTP 1.1 Server status codes -- see RFC 2048
        addStatusCodeMap(SC_CONTINUE, "Continue");
        addStatusCodeMap(SC_TEMPORARY_REDIRECT, "Temporary Redirect");
        addStatusCodeMap(SC_METHOD_NOT_ALLOWED, "Method Not Allowed");
        addStatusCodeMap(SC_CONFLICT, "Conflict");
        addStatusCodeMap(SC_PRECONDITION_FAILED, "Precondition Failed");
        addStatusCodeMap(SC_REQUEST_URI_TOO_LONG,
                "Request-URI Too Long");
        addStatusCodeMap(SC_UNSUPPORTED_MEDIA_TYPE,
                "Unsupported Media Type");
        addStatusCodeMap(SC_MULTIPLE_CHOICES, "Multiple Choices");
        addStatusCodeMap(SC_SEE_OTHER, "See Other");
        addStatusCodeMap(SC_USE_PROXY, "Use Proxy");
        addStatusCodeMap(SC_PAYMENT_REQUIRED, "Payment Required");
        addStatusCodeMap(SC_NOT_ACCEPTABLE, "Not Acceptable");
        addStatusCodeMap(SC_PROXY_AUTHENTICATION_REQUIRED,
                "Proxy Authentication Required");
        addStatusCodeMap(SC_REQUEST_TIMEOUT, "Request Timeout");

        addStatusCodeMap(SC_SWITCHING_PROTOCOLS, "Switching Protocols");
        addStatusCodeMap(SC_NON_AUTHORITATIVE_INFORMATION,
                "Non Authoritative Information");
        addStatusCodeMap(SC_RESET_CONTENT, "Reset Content");
        addStatusCodeMap(SC_PARTIAL_CONTENT, "Partial Content");
        addStatusCodeMap(SC_GATEWAY_TIMEOUT, "Gateway Timeout");
        addStatusCodeMap(SC_HTTP_VERSION_NOT_SUPPORTED,
                "Http Version Not Supported");
        addStatusCodeMap(SC_GONE, "Gone");
        addStatusCodeMap(SC_LENGTH_REQUIRED, "Length Required");
        addStatusCodeMap(SC_REQUESTED_RANGE_NOT_SATISFIABLE,
                "Requested Range Not Satisfiable");
        addStatusCodeMap(SC_EXPECTATION_FAILED, "Expectation Failed");
    }
    
    protected static void addStatusCodeMap(int code, String message) {
    	statusCodeMap.put(code, message);
    }
    
    protected static String getStatusCodeMessage(int code) {
    	return statusCodeMap.get(code);
    }

    protected boolean commited;
    protected int bufferSize;
	protected int status;
	protected int contentLength;
	protected String contentType;
	protected String characterEncoding;
	protected Locale locale;
	protected List<Cookie> cookieList;
	protected Map<String, String[]> headerMap;
	protected OutputStream originalOutputStream;
	protected boolean isUsingChunkedEncoding;
	protected ByteCountingOutputStream responseOutputStream;
	protected ByteCountingPrintWriter responsePrintWriter;
	protected boolean outputStreamRequested;
	protected boolean writerRequested;

	public HttpResponse(OutputStream outputStream) {
		this.commited = false;
		this.bufferSize = 8 * 1024;
		this.status = -1;
		this.contentLength = -1;
		this.contentType = null;
		this.characterEncoding = null;
		this.locale = null;
		this.cookieList = new ArrayList<Cookie>();
		this.headerMap = new HashMap<String, String[]>();
		this.originalOutputStream = outputStream;
		this.isUsingChunkedEncoding = false;
		this.responseOutputStream = null;
		this.responsePrintWriter = null;
		this.outputStreamRequested = false;
		this.writerRequested = false;
	}
	
	public void commit() throws UnsupportedEncodingException, IOException {
		if (commited) {
			throw new IllegalStateException(
					"Dumping state of an already commited response.");
		}
		
		if (status == -1) {
			throw new IllegalStateException("No status has been set.");
		}
		
		String headerLine;
		headerLine = HttpConstants.PROTOCOL_HTTP_1_1 +
				" " + status + " " + getStatusCodeMessage(status) + HttpConstants.CRLF;
		if (logger.isTraceEnabled()) {
			logger.trace("SENDING: " + headerLine.trim());
		}
		originalOutputStream.write(headerLine.getBytes("UTF-8"));
		headerLine = HttpConstants.HEADER_NAME_DATE + ": " +
				DateHelper.getRfc1123Format().format(new Date()) + HttpConstants.CRLF;
		if (logger.isTraceEnabled()) {
			logger.trace("SENDING: " + headerLine.trim());
		}
		originalOutputStream.write(headerLine.getBytes("UTF-8"));
		if (contentType != null) {
			headerLine = HttpConstants.HEADER_NAME_CONTENT_TYPE + ": " +
					contentType + HttpConstants.CRLF;
			if (logger.isTraceEnabled()) {
				logger.trace("SENDING: " + headerLine.trim());
			}
			originalOutputStream.write(headerLine.getBytes("UTF-8"));
		}
		if (isUsingChunkedEncoding) {
			headerLine = HttpConstants.HEADER_NAME_TRANSFER_ENCODING + ": " +
					HttpConstants.HEADER_VALUE_CHUNKED + HttpConstants.CRLF;
			if (logger.isTraceEnabled()) {
				logger.trace("SENDING: " + headerLine.trim());
			}
			originalOutputStream.write(new String(headerLine).getBytes("UTF-8"));
		} else {
			contentLength = 0;

			if (responseOutputStream != null) {
				contentLength = responseOutputStream.getByteCount();
			}
			
			if (responsePrintWriter != null) {
				contentLength = responsePrintWriter.getByteCount();
			}
			
			headerLine = HttpConstants.HEADER_NAME_CONTENT_LENGTH + ": " +
					contentLength + HttpConstants.CRLF;
			if (logger.isTraceEnabled()) {
				logger.trace("SENDING: " + headerLine.trim());
			}
			originalOutputStream.write(headerLine.getBytes("UTF-8"));
		}
		
		// Write out the cookies:
		for (Cookie cookie : cookieList) {
			headerLine = HttpConstants.HEADER_NAME_SET_COOKIE + ": " +
					cookie.serializeForResponse() + HttpConstants.CRLF;
			if (logger.isTraceEnabled()) {
				logger.trace("SENDING: " + headerLine.trim());
			}
			originalOutputStream.write(headerLine.getBytes("UTF-8"));
		}
		
		// Write out the rest of the headers:
		for (String name : headerMap.keySet()) {
			String[] values = headerMap.get(name);
			if (values == null || values.length == 0) {
				continue;
			}
			if (values.length == 1) {
				headerLine = name + ": " + values[0] + HttpConstants.CRLF;
				if (logger.isTraceEnabled()) {
					logger.trace("SENDING: " + headerLine.trim());
				}
				originalOutputStream.write(headerLine.getBytes("UTF-8"));
			} else {
				// TODO implement the case of many header values encoding
			}
		}
		
		// End of response headers.
		headerLine = HttpConstants.CRLF;
		if (logger.isTraceEnabled()) {
			logger.trace("SENDING: " + headerLine.trim());
		}
		originalOutputStream.write(headerLine.getBytes("UTF-8"));
		
		originalOutputStream.flush();
		
		commited = true;
	}
}
