package com.koutra.nodejava.registry;

import java.io.IOException;
import java.io.InputStream;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.koutra.nodejava.HttpProcessor;
import com.koutra.nodejava.HttpRequest;
import com.koutra.nodejava.HttpResponse;
import com.koutra.nodejava.NodeJava;
import com.koutra.nodejava.NodeJavaException;
import com.koutra.nodejava.StatelessHttpProcessor;
import com.koutra.nodejava.impl.SessionContainer;

public class PatternRegistry implements HttpProcessor {
	private static final Logger logger = LoggerFactory.getLogger(PatternRegistry.class);
	private static final String DEFAULT_CONFIG_FILE = "node-java.xml";
	
	private static ThreadLocal<PatternRegistry> patternRegistry =
			new ThreadLocal<PatternRegistry>();
	
	/**
	 * Extreme caution is advised when using this method, as it relies on the
	 * fact that the stack trace has gone through the HttpProcessor implementation
	 * in this class. This is generally true in the subsequent processors further
	 * down the stack, but the user is advised to make sure this is the case.
	 * @return
	 */
	public static PatternRegistry get() {
		return patternRegistry.get();
	}
	
	protected NodeJava nodeJava;
	protected Map<Pattern, StatelessHttpProcessor> patternMap =
			new LinkedHashMap<Pattern, StatelessHttpProcessor>();
	protected Transport transport = new Transport();
	protected Session session = new Session(30 * 60); // default for session timeout
	
	public PatternRegistry(String externConfigFile, NodeJava nodeJava)
			throws ParserConfigurationException,
			SAXException, IOException, ClassNotFoundException,
			InstantiationException, IllegalAccessException, NodeJavaException {
		this.nodeJava = nodeJava;
		
		// Look in the class path for a node-java.xml file and read in
		// its contents.
		String configFile = externConfigFile == null ? DEFAULT_CONFIG_FILE : externConfigFile;
		InputStream confStream = getClass().getResourceAsStream("/" + configFile);
		DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
		Document doc = dBuilder.parse(confStream);
		
		Element rootElement = doc.getDocumentElement();
		NodeList children = rootElement.getChildNodes();
		for (int i = 0; i < children.getLength(); i++) {
			Node child = children.item(i);
			if (!(child instanceof Element)) {
				continue;
			}
			Element childElement = (Element) child;
			if (childElement.getNodeName().equals("mapping")) {
				String patternString = childElement.getAttribute("pattern");
				String className = childElement.getAttribute("className");
				Pattern pattern = Pattern.compile(patternString);
				@SuppressWarnings("unchecked")
				Class<StatelessHttpProcessor> clazz =
						(Class<StatelessHttpProcessor>) Class.forName(className);
				StatelessHttpProcessor instance = clazz.newInstance();
				instance.init(nodeJava);
				patternMap.put(pattern, instance);
			} else if (childElement.getNodeName().equals("transport")) {
				String secureFlag =
						childElement.getAttribute("secure");
				String keyStorePath =
						childElement.getAttribute("keyStorePath");
				String keyStorePassword =
						childElement.getAttribute("keyStorePassword");
				String aliasPassword =
						childElement.getAttribute("aliasPassword");
				String plainServerPort =
						childElement.getAttribute("plainServerPort");
				String secureServerPort =
						childElement.getAttribute("secureServerPort");
				String processClientPoolSize =
						childElement.getAttribute("processClientPoolSize");
				String processClientTransportPoolSize =
						childElement.getAttribute("processClientTransportPoolSize");
				
				transport.setSecure("true".equals(secureFlag));
				transport.setKeyStorePath(keyStorePath);
				transport.setKeyStorePassword(keyStorePassword);
				transport.setAliasPassword(aliasPassword);
				if (plainServerPort.length() != 0) {
					transport.setPlainServerPort(
							Integer.parseInt(plainServerPort));
				}
				if (secureServerPort.length() != 0) {
					transport.setSecureServerPort(
							Integer.parseInt(secureServerPort));
				}
				if (processClientPoolSize.length() != 0) {
					transport.setProcessClientPoolSize(
							Integer.parseInt(processClientPoolSize));
				}
				if (processClientTransportPoolSize.length() != 0) {
					transport.setProcessClientTransportPoolSize(
							Integer.parseInt(processClientTransportPoolSize));
				}
			} else if (childElement.getNodeName().equals("session")) {
				String timeout = childElement.getAttribute("timeout");
				
				session.setTimeout(Integer.parseInt(timeout));
			}
		}
	}
	
	public Transport getTransport() {
		return transport;
	}
	
	public Session getSession() {
		return session;
	}

	@Override
	public void process(HttpRequest request, HttpResponse response)
			throws NodeJavaException {
		patternRegistry.set(this);
		SessionContainer.set(nodeJava.getSessionContainer());
		try {
			for (Pattern pattern : patternMap.keySet()) {
				if (pattern.matcher(request.getPath()).matches()) {
					if (logger.isTraceEnabled()) {
						logger.trace("PatternRegistry matched pattern '" +
								pattern.toString() + "' to the request for '" +
								request.getPath() + "'.");
					}
					HttpProcessor processor = patternMap.get(pattern);
					try {
						processor.process(request, response);
					} catch (Throwable t) {
						logger.error("Error processing request.", t);
					}
					return;
				}
			}
			
			logger.warn("PatternRegistry could not find a " +
					"matching pattern for the request (" + request.getPath() + ").");
			
			UnmatchedPathProcessor processor =
					new UnmatchedPathProcessor(request.getPath());
			try {
				processor.process(request, response);
			} catch (Throwable t) {
				logger.error("Error processing request.", t);
			}
		} finally {
			patternRegistry.set(null);
			SessionContainer.set(null);
		}
	}
}
