package com.koutra.nodejava.utils;

import java.io.IOException;
import java.io.OutputStream;

import org.apache.log4j.Logger;

public class ChunkedEncodingOutputStream extends ByteCountingOutputStream {
	private static final Logger logger =
			Logger.getLogger(ChunkedEncodingOutputStream.class);

	protected byte[] buffer;
	protected int currentOffset;
	
	public ChunkedEncodingOutputStream(OutputStream underlyingOutputStream,
			int bufferSize) {
		super(underlyingOutputStream);
		this.buffer = new byte[bufferSize];
		this.currentOffset = 0;
	}
	
	protected boolean wouldExceedBuffer(int numberOfBytes) {
		return currentOffset + numberOfBytes > buffer.length;
	}
	
	protected void flushBuffer() throws IOException {
		String hexSizeString = Integer.toHexString(currentOffset);
		if (logger.isTraceEnabled()) {
			logger.trace("Sending a chunked buffer of: " + hexSizeString + " bytes.");
			logger.trace(new String(buffer, 0, currentOffset));
		}
		underlyingOutputStream.write(hexSizeString.getBytes());
		underlyingOutputStream.write(HttpConstants.CRLF.getBytes());
		underlyingOutputStream.write(buffer, 0, currentOffset);
		underlyingOutputStream.write(HttpConstants.CRLF.getBytes());
		underlyingOutputStream.flush();
		currentOffset = 0;
	}
	
	public void terminateStream() throws IOException {
		if (currentOffset > 0) {
			flushBuffer();
		}
		
		// Flush the empty buffer, to signal the end of the stream.
		flushBuffer();
	}
	
	@Override
	public void flush() throws IOException {
		if (currentOffset > 0) {
			flushBuffer();
		}
		underlyingOutputStream.flush();
	}
	
	@Override
	public void write(int b) throws IOException {
		if (wouldExceedBuffer(1)) {
			flushBuffer();
		}
		buffer[currentOffset++] = (byte) b;
	}

	@Override
	public void write(byte[] b) throws IOException {
		if (wouldExceedBuffer(b.length) && currentOffset > 0) {
			flushBuffer();
		}
		if (b.length > buffer.length) {
			// We must split the byte array supplied into smaller chunks.
			// If we are in this block, then the buffer has been flushed in
			// the beginning of the function above. So we can safely assume
			// that our buffer is empty.
			int offset = 0;
			int bytesLeft = b.length;
			while (bytesLeft > buffer.length) {
				System.arraycopy(b, offset, buffer, currentOffset, buffer.length);
				currentOffset = buffer.length;
				flushBuffer();
				offset += buffer.length;
				bytesLeft -= buffer.length;
			}
			System.arraycopy(b, offset, buffer, currentOffset, bytesLeft);
			currentOffset = bytesLeft;
			return;
		}
		System.arraycopy(b, 0, buffer, currentOffset, b.length);
		currentOffset += b.length;
	}
	
	@Override
	public void write(byte[] b, int off, int len) throws IOException {
		if (wouldExceedBuffer(len) && currentOffset > 0) {
			flushBuffer();
		}
		if (len > buffer.length) {
			// We must split the byte array supplied into smaller chunks.
			// If we are in this block, then the buffer has been flushed in
			// the beginning of the function above. So we can safely assume
			// that our buffer is empty.
			int offset = off;
			int bytesLeft = len;
			while (bytesLeft > buffer.length) {
				System.arraycopy(b, offset, buffer, currentOffset, buffer.length);
				currentOffset = buffer.length;
				flushBuffer();
				offset += buffer.length;
				bytesLeft -= buffer.length;
			}
			System.arraycopy(b, offset, buffer, currentOffset, bytesLeft);
			currentOffset = bytesLeft;
			return;
		}
		System.arraycopy(b, off, buffer, currentOffset, len);
		currentOffset += len;
	}
	
}
