package com.koutra.nodejava.utils;

import java.io.IOException;
import java.io.Writer;

import org.apache.log4j.Logger;

public class ChunkedEncodingPrintWriter extends ByteCountingPrintWriter {
	private static final Logger logger =
			Logger.getLogger(ChunkedEncodingPrintWriter.class);
	
	protected char[] buffer;
	protected int currentOffset;
	protected String charsetName;
	protected Writer underlyingWriter;

	public ChunkedEncodingPrintWriter(Writer underlyingWriter,
			String charsetName, int bufferSize) {
		super(underlyingWriter, charsetName);
		this.underlyingWriter = underlyingWriter;
		this.charsetName = charsetName;
		this.buffer = new char[bufferSize];
		this.currentOffset = 0;
	}

	protected boolean wouldExceedBuffer(int numberOfChars) {
		return currentOffset + numberOfChars > buffer.length;
	}
	
	protected void flushBuffer() throws IOException {
		int byteCount = new String(buffer, 0, currentOffset).getBytes(charsetName).length;
		String hexSizeString = Integer.toHexString(byteCount);
		if (logger.isTraceEnabled()) {
			logger.trace("Sending a chunked buffer of: " + hexSizeString + " bytes.");
			logger.trace(new String(buffer, 0, currentOffset));
		}
		underlyingWriter.write(hexSizeString);
		underlyingWriter.write(HttpConstants.CRLF);
		underlyingWriter.write(buffer, 0, currentOffset);
		underlyingWriter.write(HttpConstants.CRLF);
		underlyingWriter.flush();
		currentOffset = 0;
	}
	
	public void terminateWriter() throws IOException {
		if (currentOffset > 0) {
			flushBuffer();
		}
		
		// Flush the empty buffer, to signal the end of the writer.
		flushBuffer();
	}
	
	@Override
	public void flush() {
		try {
			if (currentOffset > 0) {
				flushBuffer();
			}
			underlyingWriter.flush();
		} catch (IOException e) {
			logger.error("Unable to flush writer.", e);
			setError();
		}
	}
	
	@Override
	public void write(int c) {
		try {
			if (wouldExceedBuffer(1)) {
				flushBuffer();
			}
			buffer[currentOffset++] = (char) c;
		} catch (IOException e) {
			logger.error("Unable to write to output", e);
			setError();
		}
	}
	
	@Override
	public void write(String s) {
		try {
			if (wouldExceedBuffer(s.length()) && currentOffset > 0) {
				flushBuffer();
			}
			if (s.length() > buffer.length) {
				// We must split the char array supplied into smaller chunks.
				// If we are in this block, then the buffer has been flushed in
				// the beginning of the function above. So we can safely assume
				// that our buffer is empty.
				int offset = 0;
				int bytesLeft = s.length();
				while (bytesLeft > buffer.length) {
					s.getChars(offset, offset + buffer.length, buffer, currentOffset);
					currentOffset = buffer.length;
					flushBuffer();
					offset += buffer.length;
					bytesLeft -= buffer.length;
				}
				s.getChars(offset, offset + bytesLeft, buffer, currentOffset);
				currentOffset = bytesLeft;
				return;
			}
			s.getChars(0, s.length(), buffer, currentOffset);
			currentOffset += s.length();
		} catch (IOException e) {
			logger.error("Unable to write to output", e);
			setError();
		}
	}
	
	@Override
	public void write(String s, int off, int len) {
		try {
			if (wouldExceedBuffer(len) && currentOffset > 0) {
				flushBuffer();
			}
			if (len > buffer.length) {
				// We must split the char array supplied into smaller chunks.
				// If we are in this block, then the buffer has been flushed in
				// the beginning of the function above. So we can safely assume
				// that our buffer is empty.
				int offset = off;
				int bytesLeft = len;
				while (bytesLeft > buffer.length) {
					s.getChars(offset, offset + buffer.length, buffer, currentOffset);
					currentOffset = buffer.length;
					flushBuffer();
					offset += buffer.length;
					bytesLeft -= buffer.length;
				}
				s.getChars(offset, offset + bytesLeft, buffer, currentOffset);
				currentOffset = bytesLeft;
				return;
			}
			s.getChars(off, off + len, buffer, currentOffset);
			currentOffset += len;
		} catch (IOException e) {
			logger.error("Unable to write to output", e);
			setError();
		}
	}
	
	@Override
	public void write(char[] ac) {
		try {
			if (wouldExceedBuffer(ac.length) && currentOffset > 0) {
				flushBuffer();
			}
			if (ac.length > buffer.length) {
				// We must split the char array supplied into smaller chunks.
				// If we are in this block, then the buffer has been flushed in
				// the beginning of the function above. So we can safely assume
				// that our buffer is empty.
				int offset = 0;
				int bytesLeft = ac.length;
				while (bytesLeft > buffer.length) {
					System.arraycopy(ac, offset, buffer, currentOffset, buffer.length);
					currentOffset = buffer.length;
					flushBuffer();
					offset += buffer.length;
					bytesLeft -= buffer.length;
				}
				System.arraycopy(ac, offset, buffer, currentOffset, bytesLeft);
				currentOffset = bytesLeft;
				return;
			}
			System.arraycopy(ac, 0, buffer, currentOffset, ac.length);
			currentOffset += ac.length;
		} catch (IOException e) {
			logger.error("Unable to write to output", e);
			setError();
		}
	}
	
	@Override
	public void write(char[] buf, int off, int len) {
		try {
			if (wouldExceedBuffer(len) && currentOffset > 0) {
				flushBuffer();
			}
			if (len > buffer.length) {
				// We must split the char array supplied into smaller chunks.
				// If we are in this block, then the buffer has been flushed in
				// the beginning of the function above. So we can safely assume
				// that our buffer is empty.
				int offset = off;
				int bytesLeft = len;
				while (bytesLeft > buffer.length) {
					System.arraycopy(buf, offset, buffer, currentOffset, buffer.length);
					currentOffset = buffer.length;
					flushBuffer();
					offset += buffer.length;
					bytesLeft -= buffer.length;
				}
				System.arraycopy(buf, offset, buffer, currentOffset, bytesLeft);
				currentOffset = bytesLeft;
				return;
			}
			System.arraycopy(buf, off, buffer, currentOffset, len);
			currentOffset += len;
		} catch (IOException e) {
			logger.error("Unable to write to output", e);
			setError();
		}
	}

}
