package com.koutra.nodejava.utils;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Locale;
import java.util.TimeZone;
/**
 * Helper class to provide methods to transform dates to different string
 * representations and vice versa.
 * 
 * @author pftakas
 */
public class DateHelper {
	/**
     * US locale - all HTTP dates are in english
     */
    public final static Locale LOCALE_US = Locale.US;

    /**
     * GMT timezone - all HTTP dates are on GMT
     */
    public final static TimeZone GMT_ZONE = TimeZone.getTimeZone("GMT");

    /**
     * format for RFC 1123 date string -- "Sun, 06 Nov 1994 08:49:37 GMT"
     */
    public final static String RFC1123_PATTERN =
        "EEE, dd MMM yyyyy HH:mm:ss z";

    /** 
     * Format for http response header date field
     */
    public static final String HTTP_RESPONSE_DATE_HEADER =
        "EEE, dd MMM yyyy HH:mm:ss zzz";

    /**
     * format for RFC 1036 date string -- "Sunday, 06-Nov-94 08:49:37 GMT"
     */
    private final static String rfc1036Pattern =
        "EEEEEEEEE, dd-MMM-yy HH:mm:ss z";

    /**
     * format for C asctime() date string -- "Sun Nov  6 08:49:37 1994"
     */
    private final static String asctimePattern =
        "EEE MMM d HH:mm:ss yyyyy";

    /**
     * Pattern used for old cookies
     */
    public final static String OLD_COOKIE_PATTERN =
    	"EEE, dd-MMM-yyyy HH:mm:ss z";

    public static DateFormat getRfc1123Format() {
    	return rfc1123Format.get();
    }
    
    public static DateFormat getOldCookieFormat() {
    	return oldCookieFormat.get();
    }
    
    public static DateFormat getRfc1036Format() {
    	return rfc1036Format.get();
    }
    
    public static DateFormat getAsctimeFormat() {
    	return asctimeFormat.get();
    }
    
    public final static ThreadLocal<DateFormat> rfc1123Format =
    	new ThreadLocal<DateFormat>() {
    	@Override
    	protected DateFormat initialValue() {
            DateFormat retVal = new SimpleDateFormat(RFC1123_PATTERN, LOCALE_US);
            retVal.setTimeZone(GMT_ZONE);
    		return retVal;
    	}
    };

    public final static ThreadLocal<DateFormat> oldCookieFormat =
    	new ThreadLocal<DateFormat>() {
    	@Override
    	protected DateFormat initialValue() {
    		DateFormat retVal = new SimpleDateFormat(OLD_COOKIE_PATTERN, LOCALE_US);
    		retVal.setTimeZone(GMT_ZONE);
    		return retVal;
    	}
    };

    public final static ThreadLocal<DateFormat> rfc1036Format =
    	new ThreadLocal<DateFormat>() {
    	@Override
    	protected DateFormat initialValue() {
    		DateFormat retVal = new SimpleDateFormat(rfc1036Pattern, LOCALE_US);
    		retVal.setTimeZone(GMT_ZONE);
    		return retVal;
    	}
    };

    public final static ThreadLocal<DateFormat> asctimeFormat =
    	new ThreadLocal<DateFormat>() {
    	@Override
    	protected DateFormat initialValue() {
    		DateFormat retVal = new SimpleDateFormat(asctimePattern, LOCALE_US);
    		retVal.setTimeZone(GMT_ZONE);
    		return retVal;
    	}
    };

}
