package com.koutra.nodejava.velocity;

import java.io.IOException;

import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;

import com.koutra.nodejava.HttpRequest;
import com.koutra.nodejava.HttpResponse;
import com.koutra.nodejava.NodeJavaException;
import com.koutra.nodejava.utils.HttpConstants;

public abstract class NodeJavaVelocityContext extends VelocityContext {
	
	protected Template template;
	protected int status;
	protected String contentType;
	
	public NodeJavaVelocityContext() {
		this.template = null;
		this.status = HttpResponse.SC_OK;
		this.contentType = HttpConstants.CONTENT_TYPE_TEXT_HTML;
	}
	
	public Template getTemplate() {
		return template;
	}

	public void setTemplate(Template template) {
		this.template = template;
	}

	public int getStatus() {
		return status;
	}

	public void setStatus(int status) {
		this.status = status;
	}

	public String getContentType() {
		return contentType;
	}

	public void setContentType(String contentType) {
		this.contentType = contentType;
	}
	
	/**
	 * Send a normal response to the client. This will use the velocity
	 * template along with this context to write to the response object.
	 * 
	 * @param response the response to write to.
	 * @throws NodeJavaException Iff something goes wrong during the writting
	 * of the response.
	 */
	protected void sendNormalResponse(HttpResponse response)
			throws NodeJavaException {
		if (template == null) {
			throw new NodeJavaException("No template has been set.");
		}
		
		try {
			response.setStatus(status);
			response.setContentType(contentType);

			template.merge(this, response.getWriter());
		} catch (Exception e) {
			throw new NodeJavaException("Unable to write normal response.", e);
		}
	}
	
	/**
	 * Send a redirect response back to the client.
	 * 
	 * @param path the path to redirect to.
	 * @param response the response to write to.
	 * @throws NodeJavaException Iff something goes wrong during processing.
	 */
	protected void sendRedirect(String path, HttpResponse response)
			throws NodeJavaException {
		try {
			response.sendRedirect(path);
		} catch (IOException e) {
			throw new NodeJavaException("Unable to write redirect response.", e);
		}
	}
	
	/**
	 * Basically moves responsibility of the processing of a request to another
	 * velocity processor.
	 * 
	 * @param path the path to the new velocity context template
	 * @param request the request object
	 * @param response the response object
	 * @throws NodeJavaException Iff something goes wrong during the processing.
	 */
	protected void serverSideRedirect(String path, HttpRequest request,
			HttpResponse response) throws NodeJavaException {
		try {
			String contextClassName = VelocityProcessor.convertPathToClassName(path);
			
			@SuppressWarnings("rawtypes")
			Class javaClass = Class.forName(contextClassName);
			NodeJavaVelocityContext vc = (NodeJavaVelocityContext)
					javaClass.newInstance();
			
			vc.setTemplate(Velocity.getTemplate(path));
			
			vc.process(request, response);
		} catch (Exception e) {
			throw new NodeJavaException(
					"Unable to perform server side redirect.", e);
		}
	}

	/**
	 * Process the request. Normally after setting the context state,
	 * checking request parameters, performing any necessary actions, this
	 * will complete the cycle by a call to NodeJavaVelocityContext#sendNormalResponse
	 * method which will use the associated template to render the response.
	 * 
	 * @param request the request object.
	 * @param response the response object.
	 * @throws NodeJavaException Iff something goes wrong during the processing.
	 */
	public abstract void process(HttpRequest request, HttpResponse response)
			throws NodeJavaException;
}
